# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'loopwidget.ui'
#
# Created by: PyQt5 UI code generator 5.12.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_LoopWidget(object):
    def setupUi(self, LoopWidget):
        LoopWidget.setObjectName("LoopWidget")
        LoopWidget.resize(763, 601)
        self.horizontalLayout = QtWidgets.QHBoxLayout(LoopWidget)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.graphicsView = PlotWidget(LoopWidget)
        self.graphicsView.setObjectName("graphicsView")
        self.horizontalLayout.addWidget(self.graphicsView)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.lblDelay = QtWidgets.QLabel(LoopWidget)
        self.lblDelay.setText("")
        self.lblDelay.setObjectName("lblDelay")
        self.gridLayout.addWidget(self.lblDelay, 5, 1, 1, 1)
        self.btnPrev = QtWidgets.QPushButton(LoopWidget)
        self.btnPrev.setObjectName("btnPrev")
        self.gridLayout.addWidget(self.btnPrev, 1, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 4, 0, 1, 1)
        self.label_2 = QtWidgets.QLabel(LoopWidget)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 2, 1, 1, 1)
        self.label = QtWidgets.QLabel(LoopWidget)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 2, 0, 1, 1)
        self.lblGala = QtWidgets.QLabel(LoopWidget)
        self.lblGala.setText("")
        self.lblGala.setObjectName("lblGala")
        self.gridLayout.addWidget(self.lblGala, 8, 1, 1, 1)
        self.label_4 = QtWidgets.QLabel(LoopWidget)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 5, 0, 1, 1)
        self.lblIdx = QtWidgets.QLabel(LoopWidget)
        self.lblIdx.setAlignment(QtCore.Qt.AlignCenter)
        self.lblIdx.setObjectName("lblIdx")
        self.gridLayout.addWidget(self.lblIdx, 3, 0, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem1, 0, 0, 1, 1)
        self.label_3 = QtWidgets.QLabel(LoopWidget)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 6, 0, 1, 1)
        self.label_5 = QtWidgets.QLabel(LoopWidget)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 7, 0, 1, 1)
        self.lblBeta = QtWidgets.QLabel(LoopWidget)
        self.lblBeta.setText("")
        self.lblBeta.setObjectName("lblBeta")
        self.gridLayout.addWidget(self.lblBeta, 7, 1, 1, 1)
        self.label_7 = QtWidgets.QLabel(LoopWidget)
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 8, 0, 1, 1)
        spacerItem2 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem2, 0, 1, 1, 1)
        self.btnNext = QtWidgets.QPushButton(LoopWidget)
        self.btnNext.setObjectName("btnNext")
        self.gridLayout.addWidget(self.btnNext, 1, 1, 1, 1)
        self.lblAlpha = QtWidgets.QLabel(LoopWidget)
        self.lblAlpha.setText("")
        self.lblAlpha.setObjectName("lblAlpha")
        self.gridLayout.addWidget(self.lblAlpha, 6, 1, 1, 1)
        self.lblTot = QtWidgets.QLabel(LoopWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lblTot.sizePolicy().hasHeightForWidth())
        self.lblTot.setSizePolicy(sizePolicy)
        self.lblTot.setAlignment(QtCore.Qt.AlignCenter)
        self.lblTot.setObjectName("lblTot")
        self.gridLayout.addWidget(self.lblTot, 3, 1, 1, 1)
        spacerItem3 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem3, 4, 1, 1, 1)
        self.verticalLayout_2.addLayout(self.gridLayout)
        self.comboSeries = QtWidgets.QComboBox(LoopWidget)
        self.comboSeries.setObjectName("comboSeries")
        self.verticalLayout_2.addWidget(self.comboSeries)
        self.btnDelete = QtWidgets.QPushButton(LoopWidget)
        self.btnDelete.setObjectName("btnDelete")
        self.verticalLayout_2.addWidget(self.btnDelete)
        self.horizontalLayout.addLayout(self.verticalLayout_2)

        self.retranslateUi(LoopWidget)
        QtCore.QMetaObject.connectSlotsByName(LoopWidget)

    def retranslateUi(self, LoopWidget):
        _translate = QtCore.QCoreApplication.translate
        LoopWidget.setWindowTitle(_translate("LoopWidget", "Form"))
        self.btnPrev.setText(_translate("LoopWidget", "<"))
        self.label_2.setText(_translate("LoopWidget", "of"))
        self.label.setText(_translate("LoopWidget", "Loop #"))
        self.label_4.setText(_translate("LoopWidget", "Delay (ms)"))
        self.lblIdx.setText(_translate("LoopWidget", "0"))
        self.label_3.setText(_translate("LoopWidget", "α"))
        self.label_5.setText(_translate("LoopWidget", "β"))
        self.label_7.setText(_translate("LoopWidget", "GALA"))
        self.btnNext.setText(_translate("LoopWidget", ">"))
        self.lblTot.setText(_translate("LoopWidget", "0"))
        self.btnDelete.setText(_translate("LoopWidget", "Delete current Loop"))


from pyqtgraph import PlotWidget



from collections import namedtuple
from functools import partial

import numpy as np
import pandas as pd
import pyqtgraph as pg
from PyQt5 import QtWidgets

from graphysio import exporter, ui

Point     = namedtuple('Point', ['x', 'y'])
Cardinals = namedtuple('Cardinals', ['A', 'B', 'C'])
Angles    = namedtuple('Angles', ['alpha', 'beta', 'gala'])

class LoopWidget(ui.Ui_LoopWidget, QtWidgets.QWidget):
    def __init__(self, u, p, subsetrange, parent):
        super().__init__(parent=parent)
        self.setupUi(self)
        self.properties = {}

        self.subsetrange = subsetrange

        self.exporter = exporter.PuExporter(self, p.name())

        self.btnPrev.clicked.connect(self.prevloop)
        self.btnNext.clicked.connect(self.nextloop)
        self.btnDelete.clicked.connect(self.delloop)

        self.curidx = 0
        self.loops = []
        self.pen = p.opts['pen']
        self.graphicsView.setBackground('w')

        plotitem = self.graphicsView.getPlotItem()
        vb = plotitem.getViewBox()
        vb.setAspectLocked(lock=True, ratio=1)

        self.curveitem = pg.PlotCurveItem()
        self.scatteritem = pg.ScatterPlotItem()
        plotitem.addItem(self.scatteritem)
        plotitem.addItem(self.curveitem)

        self.initloopdata(u, p)

        if len(self.loops) > 0:
            self.lblTot.setText(str(len(self.loops)))
            self.renderloop(0)

    def initloopdata(self, u, p):
        ubegins, udurations = u.getCycleIndices(self.subsetrange)
        pbegins, pdurations = p.getCycleIndices(self.subsetrange)
        durations = map(min, zip(udurations, pdurations))
        for ubegin, pbegin, duration in zip(ubegins, pbegins, durations):
            loopu = u.series.loc[ubegin:ubegin+duration]
            loopp = p.series.loc[pbegin:pbegin+duration]
            self.loops.append(PULoop(loopu, loopp))

    def renderloop(self, idx=None):
        if idx is None:
            idx = self.curidx

        try:
            curloop = self.loops[idx]
        except IndexError:
            return

        self.lblIdx.setText(str(idx + 1))

        delay = int(curloop.offset / 1e6) # ns to ms
        self.lblDelay.setText(str(delay))

        round1 = partial(round, ndigits=1)
        alpha, beta, gala = map(round1, curloop.angles)
        self.lblAlpha.setText(str(alpha))
        self.lblBeta.setText(str(beta))
        self.lblGala.setText(str(gala))

        card = curloop.cardpoints
        cardx, cardy = zip(*card)

        self.curveitem.setData(curloop.u.values, curloop.p.values, pen=self.pen)
        self.scatteritem.setData(np.array(cardx), np.array(cardy))

    def prevloop(self):
        idx = self.curidx - 1
        if idx >= 0:
            self.curidx = idx
            self.renderloop()

    def nextloop(self):
        idx = self.curidx + 1
        if idx < len(self.loops):
            self.curidx = idx
            self.renderloop()

    def delloop(self):
        try:
            self.loops.pop(self.curidx)
        except IndexError:
            return
        if self.curidx >= len(self.loops):
            # Handle the case where we deleted the last item
            self.curidx = len(self.loops) - 1
        self.lblTot.setText(str(len(self.loops)))
        self.renderloop()

    @property
    def menu(self):
        m = {'Export' : {'&Loop Data to CSV directory' : self.exporter.exportloops}}
        return m


class PULoop(object):
    def __init__(self, u, p):
        self.__angles = None
        self.__cardpoints = None

        # Realign pressure and flow
        offset = p.index[0] - u.index[0]
        u.index += offset
        self.offset = abs(offset)

        df = pd.concat([u, p], axis=1)
        self.df = df.interpolate(method='index')

        self.u = self.df[u.name]
        self.p = self.df[p.name]

    @property
    def cardpoints(self):
        if self.__cardpoints is None:
            idxpmax = self.p.idxmax()
            idxvmax = self.u.idxmax()
            A = Point(self.u.iloc[0], self.p.iloc[0])
            B = Point(self.u.loc[idxvmax], self.p.loc[idxvmax])
            C = Point(self.u.loc[idxpmax], self.p.loc[idxpmax])
            self.__cardpoints = Cardinals(A, B, C)
        return self.__cardpoints

    @property
    def angles(self):
        if self.__angles is None:
            card = self.cardpoints
            alpha = calcangle(card.A, card.B)
            beta  = calcangle(card.A, card.B, card.C)
            gala  = calcangle(card.A, card.C)
            self.__angles = Angles(alpha, beta, gala)
        return self.__angles

def calcangle(looporigin, pointb, pointa=None):
    orig = complex(looporigin.x, looporigin.y)
    cb = complex(pointb.x, pointb.y) - orig
    if pointa is None:
        ca = complex(1, 0)
    else:
        ca = complex(pointa.x, pointa.y) - orig
    angca = np.angle(ca, deg=True)
    angcb = np.angle(cb, deg=True)
    return abs(angca - angcb)



__all__ = ['curveproperties', 'curveselect', 'cycledetect', 'datetime', 'filter', 'loopwidget', 'mainwindow', 'newplot', 'periodexport', 'poiwidget', 'setuppuloop']

from .curveproperties import  Ui_CurveProperties
from .curveselect import Ui_CurveSelection
from .cycledetect import Ui_CycleDetection
from .datetime import Ui_SetDateTime
from .filter import Ui_Filter
from .loopwidget import Ui_LoopWidget
from .mainwindow import Ui_MainWindow
from .newplot import Ui_NewPlot
from .periodexport import Ui_PeriodExport
from .poiwidget import Ui_POISelectorWidget
from .setuppuloop import Ui_SetupPULoop



import os
import itertools
from functools import partial

# Error printing
import sys
import traceback

from PyQt5 import QtCore, QtWidgets

from graphysio import dialogs, utils, ui, readdata
from graphysio.plotwidgets import TSWidget


class MainUi(ui.Ui_MainWindow, QtWidgets.QMainWindow):
    hasdata   = QtCore.pyqtSignal(object)
    haserror  = QtCore.pyqtSignal(object)
    setcoords = QtCore.pyqtSignal(float, float)

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setupUi(self)
        self.dircache = os.path.expanduser('~')

        self.tabWidget.tabCloseRequested.connect(self.closeTab)
        self.tabWidget.currentChanged.connect(self.tabChanged)

        getCLIShell = partial(utils.getshell, ui=self)
        launchNewPlot    = partial(self.launchReadData, newwidget=True)
        launchAppendPlot = partial(self.launchReadData, newwidget=False)

        mnewplot = self.menuFile.addMenu('New Plot')
        for lbl in readdata.readers:
            f = partial(launchNewPlot, datasource=lbl)
            mnewplot.addAction(f'From {lbl.capitalize()}', self.errguard(f))

        mappplot = self.menuFile.addMenu('Append to Plot')
        for lbl in readdata.readers:
            f = partial(launchAppendPlot, datasource=lbl)
            mappplot.addAction(f'From {lbl.capitalize()}', self.errguard(f))

        self.menuFile.addSeparator()
        self.menuFile.addAction('&Load plugin', self.errguard(utils.loadmodule))
        self.menuFile.addAction('Get CLI shell', self.errguard(getCLIShell))
        self.menuFile.addSeparator()
        self.menuFile.addAction('&Quit', self.close, QtCore.Qt.CTRL + QtCore.Qt.Key_Q)

        self.haserror.connect(utils.displayError)
        self.setcoords.connect(self.setCoords)

    def setCoords(self, x, y):
        dt = x / 1e6 # convert from ns to ms
        date = QtCore.QDateTime.fromMSecsSinceEpoch(dt)
        date = date.toTimeSpec(QtCore.Qt.UTC)
        timestr = date.toString("dd/MM/yyyy hh:mm:ss.zzz")
        self.lblCoords.setText(f'Time: {timestr}, Value: {y}')

    def errguard(self, f):
        # Lift exceptions to UI reported errors
        def wrapped():
            try:
                f()
            except Exception as e:
                traceback.print_exc(file=sys.stdout)
                self.haserror.emit(e)
        return wrapped

    def addTab(self, widget: QtWidgets.QWidget, tabname: str) -> None:
        tabnames = []
        for i in range(self.tabWidget.count()):
            tabnames.append(self.tabWidget.tabText(i))
        if tabname in tabnames:
            tmptabname = tabname
            # Duplicate tab name. Add a number to the end
            for i in itertools.count():
                tmptabname = f'{tabname}{i+1}'
                if tmptabname not in tabnames:
                    break
            tabname = tmptabname
        tabindex = self.tabWidget.addTab(widget, tabname)
        self.tabWidget.setCurrentIndex(tabindex)

    def closeTab(self, i) -> None:
        w = self.tabWidget.widget(i)
        self.tabWidget.removeTab(i)
        w.close()
        w.deleteLater()
        del w

    def tabChanged(self, tabid):
        destwidget = self.tabWidget.widget(tabid)
        if destwidget is None:
            return
        self.menuBar.clear()
        self.menuBar.addMenu(self.menuFile)
        for title, submenu in destwidget.menu.items():
            menu = self.menuBar.addMenu(title)
            for title, item in submenu.items():
                menu.addAction(title, item)

    def launchReadData(self, datasource, newwidget=True, cb=None):
        try:
            self.hasdata.disconnect()
        except:
            pass
        if cb is not None:
            title = "Load Data"
            self.hasdata.connect(cb)
        elif newwidget:
            title = "New Plot"
            self.hasdata.connect(self.createNewPlotWithData)
        else:
            title = "Append to Plot"
            self.hasdata.connect(self.appendToPlotWithData)

        def cb(datarequest):
            self.dircache = datarequest.folder
            self.lblStatus.setText(f'Loading... {datarequest.name}...')
            Reader = readdata.readers[datasource]
            reader = Reader(datarequest, self.hasdata, self.haserror)
            QtCore.QThreadPool.globalInstance().start(reader)

        DlgNewPlot = readdata.dlgNewPlots[datasource]
        dlgNewplot = DlgNewPlot(parent=self, title=title, directory=self.dircache)
        dlgNewplot.dlgdata.connect(cb)
        dlgNewplot.exec_()


    def appendToPlotWithData(self, plotdata, destidx=None):
        if destidx is None:
            plotwidget = self.tabWidget.currentWidget()
        else:
            plotwidget = self.tabWidget.widget(destidx)
        if plotwidget is None:
            self.haserror.emit('No plot selected.')
            return
        dorealign = dialogs.userConfirm('Timeshift new curves to make the beginnings coincide?', title='Append to plot')

        for fieldname in plotdata.data:
            newname = plotwidget.validateNewCurveName(fieldname)
            if newname is None:
                continue
            if newname != fieldname:
                plotdata.data.rename(columns={fieldname : newname}, inplace=True)

        plotwidget.appendData(plotdata, dorealign)
        plotwidget.properties['dircache'] = self.dircache
        self.lblStatus.setText("Loading... done")

    def createNewPlotWithData(self, plotdata):
        plotwidget = TSWidget(plotdata=plotdata, parent=self)
        plotwidget.properties['dircache'] = self.dircache
        self.addTab(plotwidget, plotdata.name)
        self.lblStatus.setText("Loading... done")



import pyqtgraph as pg

__all__ = ['LegendItem']

class LegendItem(pg.LegendItem):
    def __init__(self, size=None, offset=(40,5)):
        super().__init__(size, offset)

    def paint(self, p, *args):
        p.setPen(pg.mkPen(0,0,0,255))
        p.setBrush(pg.mkBrush(255,255,255,255))
        p.drawRect(self.boundingRect())

    def addItem(self, item, name):
        # Make the line width bigger
        pen = pg.mkPen(item.opts['pen'], width=2)
        newitem = pg.PlotDataItem(pen=pen)
        super().addItem(newitem, name)



from typing import Dict

import itertools
from functools import partial
from collections import namedtuple
from datetime import datetime

import numpy as np
import pandas as pd
from scipy import signal, interpolate

from graphysio.utils import truncatevecs, estimateSampleRate

Filter = namedtuple('Filter', ['name', 'parameters'])
Parameter = namedtuple('Parameter', ['description', 'request'])

class TF(object):
    def __init__(self, num, den, name=''):
        self.num  = num
        self.den  = den
        self.name = name

    def discretize(self, samplerate):
        systf = (self.num, self.den)
        (dnum, dden, _) = signal.cont2discrete(systf, 1 / samplerate)
        return (np.squeeze(dnum), np.squeeze(dden))

interpkind = ['linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic']
Filters = {
    'Lowpass filter' : Filter(name='lowpass', parameters=[Parameter('Cutoff frequency (Hz)', int), Parameter('Filter order', int)]),
    'Filter ventilation' : Filter(name='ventilation', parameters=[]),
    'Savitzky-Golay' : Filter(name='savgol', parameters=[Parameter('Window duration', 'time'), Parameter('Polynomial order', int)]),
    'Interpolate' : Filter(name='interp', parameters=[Parameter('New sampling rate (Hz)', float), Parameter('Interpolation type', interpkind)]),
    'Doppler cut' : Filter(name='dopplercut', parameters=[Parameter('Minimum value', int)]),
    'Fill NaNs' : Filter(name='fillnan', parameters=[]),
    'Differentiate' : Filter(name='diff', parameters=[Parameter('Order', int)]),
    'Integrate' : Filter(name='integrate', parameters=[Parameter('Window duration', 'time')]),
    'Lag' : Filter(name='lag', parameters=[Parameter('Amount (s)', float)]),
    'Normalize' : Filter(name='norm1', parameters=[]),
    'Set start date/time' : Filter(name='setdatetime', parameters=[Parameter('DateTime', datetime)]),
    'Tolerant Normalize' : Filter(name='norm2', parameters=[]),
    'Enter expression (variable = x)' : Filter(name='expression', parameters=[Parameter('Expression', str)]),
    'Pressure scale' : Filter(name='pscale', parameters=[Parameter('Systole', int), Parameter('Diastole', int), Parameter('Mean', int)]),
    'Strided Moving average' : Filter(name='sma', parameters=[Parameter('Window duration', 'time')]),
    'Affine scale' : Filter(name='affine', parameters=[Parameter('Scale factor', float), Parameter('Translation factor', float)])
}

FeetFilters = {
    'Short cycles' : Filter(name='shortcycles', parameters=[Parameter('Minimum duration', 'time')])
}

TFs: Dict[str, TF] = {}
def updateTFs():
    tflist = list(TFs.keys())
    if not tflist:
        return
    Filters['Transfer function'] = Filter(name='tf', parameters=[Parameter('Transfer function', tflist)])

def norm1(series, samplerate, parameters):
    series -= np.mean(series)
    series /= np.max(series) - np.min(series)
    newname = f'{series.name}-norm'
    return (series.rename(newname), samplerate)

def norm2(series, samplerate, parameters):
    series -= np.mean(series)
    series /= np.std(series)
    newname = f'{series.name}-norm'
    return (series.rename(newname), samplerate)

from sympy import lambdify
from sympy.abc import x
from sympy.parsing.sympy_parser import parse_expr
def expression(series, samplerate, parameters):
    express, = parameters
    expr = parse_expr(express)
    f = lambdify(x, expr, 'numpy')
    filtered = f(series.values)
    newname = f'{series.name}-filtered'
    newseries = pd.Series(filtered, index=series.index, name=newname)
    return (newseries, samplerate)

def setdatetime(series, samplerate, parameters):
    timestamp, = parameters
    newseries = series.copy()
    diff = timestamp - newseries.index[0]
    newseries.index += diff
    newseries = newseries.rename(f'{series.name}-{timestamp}')
    return (newseries, samplerate)

def sma(series, samplerate, parameters):
    window_s, = parameters
    winsize = int(window_s * samplerate)
    nwindows = int(series.size / winsize)
    stacked = np.vstack([series.index.values, series.values])
    windows = np.lib.stride_tricks.as_strided(stacked, shape=(2, nwindows, winsize))
    result = np.apply_along_axis(np.mean, 2, windows)
    newname = f'{series.name}-sma{window_s}s'
    newseries = pd.Series(result[1], index=result[0], name=newname)
    newsamplerate = samplerate / winsize
    return (newseries, newsamplerate)

def savgol(series, samplerate, parameters):
    window, order = parameters
    window = np.floor(window * samplerate)
    if not window % 2:
        # window is even, we need odd
        window += 1
    filtered = signal.savgol_filter(series.values, int(window), order)
    newname = f'{series.name}-filtered'
    newseries = pd.Series(filtered, index=series.index, name=newname)
    return (newseries, samplerate)

def lag(series, samplerate, parameters):
    timedelta, = parameters
    nindex = series.index.values + timedelta * 1e9
    newname = f'{series.name}-{timedelta}s'
    newseries = pd.Series(series.values, index=nindex, name=newname)
    return (newseries, samplerate)

def affine(series, samplerate, parameters):
    a, b = parameters
    filtered = a * series + b
    newname = f'{series.name}-affine-{a}-{b}'
    newseries = pd.Series(filtered, index=series.index, name=newname)
    return (newseries, samplerate)

def tf(series, samplerate, parameters):
    filtname, = parameters
    tf = TFs[filtname]
    b, a = tf.discretize(samplerate)
    filtered = signal.lfilter(b, a, series)
    newname = f'{series.name}-{tf.name}'
    newseries = pd.Series(filtered, index=series.index, name=newname)
    return (newseries, samplerate)

def lowpass(series, samplerate, parameters):
    Fc, order = parameters
    Wn = Fc * 2 / samplerate
    b, a = signal.butter(order, Wn)
    filtered = signal.lfilter(b, a, series)
    newname = f'{series.name}-lp{Fc}'
    newseries = pd.Series(filtered, index=series.index, name=newname)
    return (newseries, samplerate)

def ventilation(series, samplerate, parameters):
    order = 12
    # Filter between 10 and 20 per minute
    bornes_hz = np.array([10, 20]) / 60
    Wn = [Fc * 2 / samplerate for Fc in bornes_hz]
    b, a = signal.butter(order, Wn, btype='bandstop')
    filtered = signal.filtfilt(b, a, series.values)
    newname = f'{series.name}-novent'
    newseries = pd.Series(filtered, index=series.index, name=newname)
    return (newseries, samplerate)

def interp(series, samplerate, parameters):
    newsamplerate, method = parameters
    oldidx = series.index
    f = interpolate.interp1d(oldidx, series.values, kind=method)
    step = 1e9 / newsamplerate # 1e9 to convert Hz to ns
    newidx = np.arange(oldidx[0], oldidx[-1], step, dtype=np.int64)
    resampled = f(newidx)
    newname = f'{series.name}-{newsamplerate}Hz'
    newseries = pd.Series(resampled, index=newidx, name=newname)
    return (newseries, newsamplerate)

def dopplercut(series, samplerate, parameters):
    minvel, = parameters
    notlow, = (series < minvel).nonzero()
    newname = f'{series.name}-{minvel}+'
    newseries = series.rename(newname)
    newseries.iloc[notlow] = 0
    return (newseries, samplerate)

def integrate(series, samplerate, parameters):
    duration, = parameters
    window = int(np.floor(duration * samplerate))
    integrated = series.rolling(window, center=True).sum()
    newname = f'{series.name}-integrate{duration}s'
    newseries = integrated.rename(newname)
    return (newseries, samplerate)

def diff(series, samplerate, parameters):
    order, = parameters
    diffed = np.diff(series.values, order)
    newname = f'{series.name}-diff{order}'
    newseries = pd.Series(diffed, index=series.index[order:], name=newname)
    return (newseries, samplerate)

def fillnan(series, samplerate, parameters):
    newseries = series.interpolate(method='index', limit_direction='both')
    newseries = newseries.rename(f'{series.name}-nona')
    return (newseries, samplerate)

def pscale(series, samplerate, parameters):
    sbp, dbp, mbp = parameters
    detrend = series - series.mean()
    scaled = detrend * (sbp - dbp) / (series.max() - series.min())
    newseries = scaled + mbp
    newseries = newseries.rename(f'{series.name}-pscale')
    return (newseries, samplerate)

filtfuncs = {'savgol'     : savgol,
             'affine'     : affine,
             'lag'        : lag,
             'tf'         : tf,
             'sma'        : sma,
             'lowpass'    : lowpass,
             'ventilation': ventilation,
             'interp'     : interp,
             'dopplercut' : dopplercut,
             'integrate'  : integrate,
             'diff'       : diff,
             'pscale'     : pscale,
             'norm1'      : norm1,
             'norm2'      : norm2,
             'expression' : expression,
             'setdatetime' : setdatetime,
             'fillnan'    : fillnan}

def filter(curve, filtname, paramgetter):
    samplerate = curve.samplerate
    series = curve.series
    filt = Filters[filtname]
    parameters = map(paramgetter, filt.parameters)
    return filtfuncs[filt.name](series, samplerate, parameters)

def filterFeet(starts, stops, filtname, paramgetter):
    filt = FeetFilters[filtname]
    parameters = map(paramgetter, filt.parameters)

    if filt.name == 'shortcycles':
        if len(stops) < 1:
            # No stop information
            raise ValueError('No stop feet')
        minDuration, = parameters
        minCycleLength = minDuration * 1e9 # Transform ms to ns
        cycleDurations = (stop - start for start, stop in zip(starts, stops))
        boolidx = [d >= minCycleLength for d in cycleDurations]
        starts, stops, boolidx = truncatevecs([starts, stops, boolidx])
        newstarts = starts[boolidx]
        newstops = stops[boolidx]
    else:
        errmsg = f'Unknown filter: {filtname}'
        raise ValueError(errmsg)
    return (newstarts, newstops)



import string
from functools import partial

import numpy as np
import pandas as pd
import pyqtgraph as pg
import sympy

from pyqtgraph.Qt import QtGui, QtCore

from graphysio import exporter, utils, dialogs, transformations
from graphysio.types import CycleId, Parameter, PlotData
from graphysio.algorithms import filters

from graphysio.plotwidgets import PlotWidget, LoopWidget, POISelectorWidget, SpectrogramWidget
from graphysio.plotwidgets.curves import CurveItemWithPOI


class TSWidget(PlotWidget):
    @staticmethod
    def mouseMoved(self, evt):
        pos = evt[0]  ## using signal proxy turns original arguments into a tuple
        if self.sceneBoundingRect().contains(pos):
            mousePoint = self.getViewBox().mapSceneToView(pos)
            self.parent.setcoords.emit(mousePoint.x(), mousePoint.y())

    def __init__(self, plotdata, parent=None):
        super().__init__(plotdata.name, parent=parent)
        self.exporter = exporter.TsExporter(self, plotdata.name)
        self.appendData(plotdata)
        mouseMoved = partial(self.mouseMoved, self)
        self.sigproxy = pg.SignalProxy(self.scene().sigMouseMoved, rateLimit=60, slot=mouseMoved)

    def filterCurve(self, oldcurve, filtername, asnew=False):
        newseries, newsamplerate = filters.filter(oldcurve, filtername, dialogs.askUserValue)
        if asnew:
            newname = self.validateNewCurveName(newseries.name)
            if newname != newseries.name:
                newseries = newseries.rename(newname)
            newcurve = self.addSeriesAsCurve(series=newseries)
            newcurve.samplerate = newsamplerate
        else:
            newseries = newseries.rename(oldcurve.series.name)
            newseries = newseries.dropna()
            if len(newseries) < 1:
                return
            newseries = newseries.groupby(newseries.index).mean()
            oldcurve.clear()
            oldcurve.series = newseries
            oldcurve.samplerate = newsamplerate
            oldcurve.render()

    def filterFeet(self, curve, filtername, asnew=False):
        feetdict = curve.feetitem.indices
        oldstarts = feetdict['start']
        oldstops = feetdict['stop']
        starts, stops = filters.filterFeet(oldstarts, oldstops, filtername, dialogs.askUserValue)
        feetdict['start'] = starts
        feetdict['stop'] = stops
        curve.feetitem.render()

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_Delete:
            for curve in self.curves.values():
                curve.feetitem.removeSelection()

    # Menu Curves
    def showCurveList(self):
        dlgCurveSelection = dialogs.DlgCurveSelection(parent=self.parent, visible=self.curves.values(), hidden=self.hiddenitems)
        if not dlgCurveSelection.exec_():
            return
        visible, invisible = dlgCurveSelection.result
        newvisible = [item for item in visible if item not in self.curves.values()]
        newinvisible = [item for item in invisible if item not in self.hiddenitems]
        for item in newvisible:
            self.addCurve(item)
            if item in self.hiddenitems:
                self.hiddenitems.remove(item)
        for item in newinvisible:
            self.removeCurve(item)
            if item not in self.hiddenitems:
                self.hiddenitems.append(item)

    def launchCycleDetection(self):
        dlgCycles = dialogs.DlgCycleDetection(parent=self.parent)

        def cb(choices):
            for curvename, choice in choices.items():
                curve = self.curves[curvename]
                curve.addFeet(CycleId(choice))

        dlgCycles.dlgdata.connect(cb)
        dlgCycles.exec_()

    def launchFilter(self, filterfeet):
        dlgFilter = dialogs.DlgFilter(parent=self.parent, filterfeet=filterfeet)

        def cb(result):
            createnew, curvechoices = result
            if filterfeet:
                filterfunc = self.filterFeet
            else:
                filterfunc = self.filterCurve
            for curvename, choice in curvechoices.items():
                if choice == 'None':
                    continue
                curve = self.curves[curvename]
                filterfunc(curve, choice, asnew=createnew)

        dlgFilter.dlgdata.connect(cb)
        dlgFilter.exec_()

    def setDateTime(self):
        if len(self.curves) < 1:
            return
        sortedcurves = sorted(self.curves.values(), key=lambda curve: curve.series.index[0])
        fstcurve = sortedcurves[0]
        curtimestamp = fstcurve.series.index[0]
        dlg = dialogs.DlgSetDateTime(prevdatetime=curtimestamp)
        isok = dlg.exec_()
        if not isok:
            return
        newtimestamp = dlg.result
        offset = newtimestamp - curtimestamp
        for curve in self.curves.values():
            curve.series.index += offset
            curve.render()

    def launchTransformation(self):
        param = Parameter("Choose Transformation", list(transformations.Transformations.keys()))
        qresult = dialogs.askUserValue(param)
        if qresult is None:
            return
        trans = transformations.Transformations[qresult]
        for curve in trans(self):
            self.addCurve(curve)

    # Menu Plot
    def launchPOIWidget(self):
        curvenames = list(self.curves.keys())
        q = Parameter('Select Curve', curvenames)
        curvename = dialogs.askUserValue(q)
        if not curvename:
            return
        curve = self.curves[curvename]
        poiselector = POISelectorWidget(curve.series, parent=self.parent)
        self.parent.addTab(poiselector, curve.name())

    def launchSpectrogram(self):
        curvenames = list(self.curves.keys())
        q = Parameter('Select Curve', curvenames)
        curvename = dialogs.askUserValue(q)
        q = Parameter('Enter time window', 'time')
        window = dialogs.askUserValue(q)
        if not curvename:
            return
        curve = self.curves[curvename]
        spectro = SpectrogramWidget(curve.series, curve.samplerate, window, parent=self.parent)
        self.parent.addTab(spectro, curve.name())

    def launchCurveAlgebra(self):
        curvecorr = {n : c for n,c in zip(list(string.ascii_lowercase)
                                         ,list(self.curves.keys()))}
        dlgCurveAlgebra = dialogs.DlgCurveAlgebra(self, curvecorr)
        if not dlgCurveAlgebra.exec_():
            return
        formula = dlgCurveAlgebra.result
        expr = sympy.sympify(formula)
        symbols = list(expr.free_symbols)
        schar = sorted([str(x) for x in symbols])
        curvenames = [curvecorr[x] for x in schar]
        sers = [self.curves[c].series for c in curvenames]
        if len(sers) < 1:
            # Scalar
            return
        df = pd.concat(sers, axis=1, keys=[s.name for s in sers])
        df = df.dropna(how='all').interpolate()
        args = [df[c].values for c in curvenames]
        l = sympy.lambdify(symbols, expr, 'numpy')
        newvals = l(*args)
        newname = self.validateNewCurveName(formula, True)
        newseries = pd.Series(newvals, index=df.index, name=newname, dtype='double')
        self.addSeriesAsCurve(newseries)


    # Menu Selection
    def launchNewPlotFromSelection(self):
        xmin, xmax = self.vbrange
        sers = []
        for c in self.curves.values():
            series = c.series.loc[xmin:xmax]
            sers.append(series)
        df = pd.concat(sers, axis=1, keys=[s.name for s in sers])
        newname = f'{self.name}-sub'
        plotdata = PlotData(data=df, name=newname)
        self.parent.createNewPlotWithData(plotdata)

    def launchAppendToPlotFromSelection(self):
        tabWidget = self.parent.tabWidget
        ntabs = tabWidget.count()
        tabdict = {tabWidget.tabText(idx):idx for idx in range(ntabs)}
        desttabname, ok = QtGui.QInputDialog.getItem(self,
                                                     'Select destination',
                                                     'Destination plot',
                                                     list(tabdict.keys()),
                                                     editable=False)
        if not ok:
            return
        xmin, xmax = self.vbrange
        sers = []
        for c in self.curves.values():
            series = c.series.loc[xmin:xmax]
            sers.append(series)
        df = pd.concat(sers, axis=1, keys=[s.name for s in sers])
        newname = f'{self.name}-sub'
        plotdata = PlotData(data=df, name=newname)
        self.parent.appendToPlotWithData(plotdata, destidx=tabdict[desttabname])

    def launchLoop(self):
        dlgSetupPU = dialogs.DlgSetupPULoop(self, parent=self.parent)

        def cb(result):
            uname, pname = result
            u = self.curves[uname]
            p = self.curves[pname]
            loopwidget = LoopWidget(u, p, self.vbrange, parent=self.parent)
            newname = f'{self.name}-{p.name()}-loops'
            self.parent.addTab(loopwidget, newname)

        dlgSetupPU.dlgdata.connect(cb)
        dlgSetupPU.exec_()

    @property
    def menu(self):
        mcurves = {'Visible Curves'  : self.showCurveList
                  ,'Cycle Detection' : self.launchCycleDetection
                  ,'Filter Curve'    : partial(self.launchFilter, filterfeet=False)
                  ,'Filter Feet'     : partial(self.launchFilter, filterfeet=True)
                  ,'Transformation'  : self.launchTransformation
                  ,'Set Date / Time' : self.setDateTime
                  }
        mselect = {'As new plot'          : self.launchNewPlotFromSelection
                  ,'Append to other plot' : self.launchAppendToPlotFromSelection
                  ,'Generate PU-Loops'    : self.launchLoop
                  }
        mplot = {'POI Selector'  : self.launchPOIWidget
                ,'Spectrogram'   : self.launchSpectrogram
                ,'Curve Algebra' : self.launchCurveAlgebra}
        mexport = {'Series to CSV'           : self.exporter.seriestocsv
                  ,'Time info to CSV'        : self.exporter.periodstocsv
                  ,'Cycle info to CSV'       : self.exporter.cyclepointstocsv
                  ,'Cycles to CSV directory' : self.exporter.cyclestocsv
                  }
        m = {'Curves' : mcurves
            ,'Plot'   : mplot
            ,'Seletion' : mselect
            ,'Export' : mexport
            }
        return m



import os
from functools import partial

import numpy as np
import pandas as pd

from pyqtgraph.Qt import QtGui, QtCore, QtWidgets

from graphysio import types, ui


class ParquetReader(QtCore.QRunnable):
    def __init__(self, datarequest, sigdata, sigerror):
        super().__init__()
        self.filepath = datarequest.filepath
        self.sigdata = sigdata
        self.sigerror = sigerror

    def run(self) -> None:
        try:
            data = self.getdata()
        except Exception as e:
            self.sigerror.emit(e)
        else:
            self.sigdata.emit(data)

    def getdata(self) -> types.PlotData:
        data = pd.read_parquet(self.filepath)

        data = data.dropna(axis='columns', how='all')
        data = data.sort_index()
        data.index = data.index.astype(np.int64)

        plotdata = types.PlotData(data=data, filepath=self.filepath)
        return plotdata


class DataRequest():
    def __init__(self, filepath=""):
        self.filepath = filepath

    @property
    def name(self):
        name, _ = os.path.splitext(os.path.basename(self.filepath))
        return name

    @property
    def folder(self):
        folder = os.path.dirname(self.filepath)
        return folder


class DlgNewPlotParquet(QtCore.QObject):
    dlgdata = QtCore.pyqtSignal(object)

    def __init__(self, parent=None, title="New Plot", directory=""):
        super().__init__(parent=parent)
        self.datarequest = DataRequest()
        self.dircache = directory

    def exec_(self):
        filepath = QtGui.QFileDialog.getOpenFileName(parent = None,
                                                     caption = "Open Parquet file",
                                                     directory = self.dircache)
        # PyQt5 API change
        if type(filepath) is not str:
            filepath = filepath[0]

        if filepath:
            self.datarequest.filepath = filepath
            self.dlgdata.emit(self.datarequest)
    



from enum import Enum
from functools import partial

import pyqtgraph as pg

from PyQt5 import QtWidgets, QtGui

from graphysio import exporter, ui
from graphysio.plotwidgets import PlotWidget
from graphysio.algorithms.filters import savgol
from graphysio.algorithms.waveform import findPOIGreedy

class FixIndex(Enum):
    disabled = 'Disabled'
    minimum  = 'Local Minimum'
    sndderiv = '2nd derivative peak'


class POISelectorWidget(ui.Ui_POISelectorWidget, QtWidgets.QWidget):
    @staticmethod
    def buttonClicked(self, qbutton):
        fixval = FixIndex(qbutton.text())
        self.poiselectorwidget.fixvalue = fixval

    def __init__(self, series, parent):
        super().__init__(parent=parent)
        self.setupUi(self)
        self.parent = parent

        self.poiselectorwidget = POISelectorPlot(series, parent=self)
        self.horizontalLayout.addWidget(self.poiselectorwidget)

        buttonClicked = partial(self.buttonClicked, self)
        self.buttonGroup.buttonClicked.connect(buttonClicked)

    @property
    def menu(self):
        m = {'Plot'   : {'&Import POIs' : partial(self.parent.launchReadData, cb=self.poiselectorwidget.loadPOI)}
            ,'Export' : {'&POI to CSV'  : self.poiselectorwidget.exporter.poitocsv}
            }
        return m


class POISelectorPlot(PlotWidget):
    pointkey = 'point'

    @staticmethod
    def mouseMoved(self, evt):
        pos = evt[0]  ## using signal proxy turns original arguments into a tuple
        if self.sceneBoundingRect().contains(pos):
            mousePoint = self.getViewBox().mapSceneToView(pos)
            self.vLine.setPos(mousePoint.x())

    @staticmethod
    def clicked(self, evt):
        button = evt.button()
        pos = self.vLine.value()
        if button == 1:
            correctedpos = self.fixpos(pos)
            self.curve.feetitem.addPointsByLocation(self.pointkey, [correctedpos])
        elif button == 2:
            self.curve.feetitem.removePointsByLocation(self.pointkey, [pos])

    def __init__(self, series, parent=None):
        super().__init__(name=series.name, parent=parent)
        self.vb.setMouseMode(self.vb.PanMode)
        self.setMenuEnabled(False)

        self.__sndderiv = None
        self.fixvalue = FixIndex.disabled

        self.curve = self.addSeriesAsCurve(series)
        self.exporter = exporter.POIExporter(self, self.name)

        pen = pg.mkPen('k', width=2)
        self.vLine = pg.InfiniteLine(angle=90, movable=False, pen=pen)
        self.addItem(self.vLine, ignoreBounds=True)

        mouseMoved = partial(self.mouseMoved, self)
        self.sigproxy = pg.SignalProxy(self.scene().sigMouseMoved, rateLimit=60, slot=mouseMoved)
        clicked = partial(self.clicked, self)
        self.scene().sigMouseClicked.connect(clicked)

    def loadPOI(self, plotdata):
        columnname, ok = QtGui.QInputDialog.getItem(self,
                                                    'Select POI series',
                                                    'Load POI',
                                                    plotdata.data.columns,
                                                    editable=False)
        if not ok:
            return
        poiseries = plotdata.data[columnname].dropna().index
        self.curve.feetitem.addPointsByLocation(self.pointkey, poiseries)

    def fixpos(self, pos):
        # Need to go through get_loc, otherwise strange things happen
        correctedpos = pos
        if self.fixvalue is FixIndex.minimum:
            xmin, xmax = self.vbrange
            s = self.curve.series.loc[xmin:xmax]
            posprop = findPOIGreedy(s, pos, 'min')
            fixedposloc = s.index.get_loc(posprop, method='nearest')
            correctedpos = s.index[fixedposloc]
        elif self.fixvalue is FixIndex.sndderiv:
            xmin, xmax = self.vbrange
            s = self.sndderiv.loc[xmin:xmax]
            posprop = findPOIGreedy(s, pos, 'max')
            fixedposloc = s.index.get_loc(posprop, method='nearest')
            correctedpos = s.index[fixedposloc]
        else:
            xmin, xmax = self.vbrange
            s = self.curve.series.loc[xmin:xmax]
            fixedposloc = s.index.get_loc(pos, method='nearest')
            correctedpos = s.index[fixedposloc]
        return correctedpos

    @property
    def sndderiv(self):
        if self.__sndderiv is None:
            sndderiv = self.curve.series.diff().diff().iloc[1:]
            sndderiv, _ = savgol(sndderiv, self.curve.samplerate, (.16, 2))
            self.__sndderiv = sndderiv.dropna()
        return self.__sndderiv



from typing import List

import pandas as pd

from graphysio.dialogs import askUserValue
from graphysio.types import Parameter
from graphysio.plotwidgets import PlotWidget
from graphysio.plotwidgets.curves import CurveItemWithPOI

def perfusionindex(plotwidget: PlotWidget) -> List[CurveItemWithPOI]:
    curvenames = list(plotwidget.curves.keys())
    q = Parameter('Select Curve', curvenames)
    curvename = askUserValue(q)
    curve = plotwidget.curves[curvename]
    if 'start' not in curve.feetitem.indices or curve.feetitem.indices['start'].size < 1:
        raise ValueError('No start information for curve')

    wave = curve.series
    starts = curve.getFeetPoints('start')
    df = pd.concat([wave, starts], axis=1)
    df = df.interpolate(method='index')

    begins, durations = curve.getCycleIndices()
    pivalues = []
    for begin, duration in zip(begins, durations):
        cycle = df.loc[begin:begin+duration]
        auctot = cycle[wave.name].sum()
        aucbase = cycle[starts.name].sum()
        aucpulse = auctot - aucbase
        pi = aucpulse / auctot
        pivalues.append(pi)

    piseries = pd.Series(pivalues, index=begins)
    piseries.name = f'{wave.name}-perf'

    newcurve = CurveItemWithPOI(parent=plotwidget, series=piseries, pen=plotwidget.getPen())
    return [newcurve]

def feettocurve(plotwidget: PlotWidget) -> List[CurveItemWithPOI]:
    feetitemhash = {}
    for curve in plotwidget.curves.values():
        feetitemhash.update({f'{curve.name()}-{feetname}' : (curve, feetname) for feetname in curve.feetitem.indices.keys()})
    param = Parameter("Choose points to create curve", list(feetitemhash.keys()))
    qresult = askUserValue(param)
    curve, feetname = feetitemhash[qresult]

    newseries = curve.getFeetPoints(feetname)
    newseries.name = qresult

    newcurve = CurveItemWithPOI(parent=plotwidget, series=newseries, pen=plotwidget.getPen())
    return [newcurve]

Transformations = {'Perfusion Index' : perfusionindex,
                   'Feet to Curve'   : feettocurve}



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'cycledetect.ui'
#
# Created by: PyQt5 UI code generator 5.12.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_CycleDetection(object):
    def setupUi(self, CycleDetection):
        CycleDetection.setObjectName("CycleDetection")
        CycleDetection.resize(387, 371)
        self.verticalLayout = QtWidgets.QVBoxLayout(CycleDetection)
        self.verticalLayout.setObjectName("verticalLayout")
        self.table = QtWidgets.QTableWidget(CycleDetection)
        self.table.setColumnCount(2)
        self.table.setObjectName("table")
        self.table.setRowCount(0)
        self.verticalLayout.addWidget(self.table)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.okButton = QtWidgets.QPushButton(CycleDetection)
        self.okButton.setObjectName("okButton")
        self.horizontalLayout.addWidget(self.okButton)
        self.cancelButton = QtWidgets.QPushButton(CycleDetection)
        self.cancelButton.setObjectName("cancelButton")
        self.horizontalLayout.addWidget(self.cancelButton)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(CycleDetection)
        QtCore.QMetaObject.connectSlotsByName(CycleDetection)

    def retranslateUi(self, CycleDetection):
        _translate = QtCore.QCoreApplication.translate
        CycleDetection.setWindowTitle(_translate("CycleDetection", "Cycle Detection"))
        self.okButton.setText(_translate("CycleDetection", "Ok"))
        self.cancelButton.setText(_translate("CycleDetection", "Cancel"))





import os
from enum import Enum

from graphysio.algorithms.filters import Filter, Parameter

class FootType(Enum):
    start    = 'start'
    stop     = 'stop'
    diastole = 'diastole'
    systole  = 'systole'
    dicrotic = 'dicrotic'

class CycleId(Enum):
    none     = 'None'
    foot     = 'Pressure foot'
    pressure = 'Pressure Full'
    velocity = 'Velocity'

class CsvRequest():
    def __init__(self, filepath  = "",
                       seperator = ",",
                       decimal   = ".",
                       dtfield  = None,
                       yfields = [],
                       datetime_format = "%Y-%m-%d %H:%M:%S,%f",
                       droplines = 0,
                       generatex = False,
                       timezone  = 'UTC',
                       encoding  = 'latin1',
                       samplerate = None):
        self.filepath = filepath
        self.seperator = seperator
        self.decimal = decimal
        self.dtfield = dtfield
        self.yfields = yfields
        self.datetime_format = datetime_format
        self.droplines = droplines
        self.generatex = generatex
        self.encoding = encoding
        self.timezone = timezone
        self.samplerate = samplerate

    @property
    def fields(self):
        dtfields = [] if self.dtfield is None else [self.dtfield]
        return dtfields + self.yfields

    @property
    def name(self):
        name, _ = os.path.splitext(os.path.basename(self.filepath))
        return name

    @property
    def folder(self):
        folder = os.path.dirname(self.filepath)
        return folder

class PlotData():
    def __init__(self, data = [],
                       filepath = "",
                       name = None):
        self.data = data
        self.filepath = filepath
        self._name = name

    @property
    def name(self):
        if self._name is not None:
            name = self._name
        else:
            name, _ = os.path.splitext(os.path.basename(self.filepath))
        return name

    @property
    def folder(self):
        folder = os.path.dirname(self.filepath)
        return folder



#!/usr/bin/env python3

import sys
from pyqtgraph.Qt import QtGui
from graphysio.mainui import MainUi

if __name__ == '__main__':
    app = QtGui.QApplication(sys.argv)

    winmain = MainUi()
    winmain.show()

    status = app.exec_()
    sys.exit(status)



from functools import partial

import numpy as np
import pandas as pd

import pyqtgraph as pg
from pyqtgraph.Qt import QtGui, QtCore, QtWidgets

from graphysio import dialogs
from graphysio.types import Parameter, PlotData

class SpectrogramWidget(QtWidgets.QWidget):
    def __init__(self, series, Fs, s_chunklen, parent=None):
        super().__init__(parent=parent)
        self.spectro = SpectrogramPlotWidget(series, Fs, s_chunklen, parent)
        self.loslider = QtWidgets.QSlider()
        self.hislider = QtWidgets.QSlider()
        layout = QtWidgets.QHBoxLayout()

        lolayout = QtWidgets.QVBoxLayout()
        lothres = QtWidgets.QLabel('Low')
        lolayout.addWidget(lothres)
        lolayout.addWidget(self.loslider)

        hilayout = QtWidgets.QVBoxLayout()
        hithres = QtWidgets.QLabel('High')
        hilayout.addWidget(hithres)
        hilayout.addWidget(self.hislider)

        layout.addWidget(self.spectro)
        layout.addWidget(lolayout)
        layout.addWidget(hilayout)
        self.setLayout(layout)

    @property
    def menu(self):
        mplot = {'Extract SEF' : self.spectro.launchSEFExtract}
        m = {'Plot': mplot}
        return m

class SpectroTimeAxisItem(pg.AxisItem):
    def __init__(self, initvalue, samplerate, chunksize, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.initvalue = initvalue / 1e6   # ns to ms
        self.samplerate = samplerate
        self.chunksize = chunksize

    def tickStrings(self, values, scale, spacing):
        ret = []
        value_to_time = self.chunksize / self.samplerate + self.initvalue
        for value in values:
            value = 1e3*value # s to ms
            value = value * value_to_time
            date = QtCore.QDateTime.fromMSecsSinceEpoch(value)
            date = date.toTimeSpec(QtCore.Qt.UTC)
            datestr = date.toString("dd/MM/yyyy\nhh:mm:ss.zzz")
            ret.append(datestr)
        return ret

#To set:
# levels
# color gradient
class SpectrogramPlotWidget(pg.PlotWidget):
    def __init__(self, series, Fs, s_chunklen, parent=None):
        # s_chunklen in seconds
        self.name = series.name
        self.parent = parent
        self.origidx = series.index.values

        self.fs = Fs
        self.data = series.values
        self.chunksize = int(s_chunklen * Fs)
        self.win = np.hanning(self.chunksize)

        axisItem = SpectroTimeAxisItem(
            initvalue = self.origidx[0],
            samplerate = self.fs,
            chunksize = self.chunksize,
            orientation = 'bottom'
        )
        axisItems = {'bottom': axisItem}
        super().__init__(parent=self.parent, axisItems=axisItems)

        self.img = pg.ImageItem()
        self.addItem(self.img)
        self.setLabel('left', 'Frequency', units='Hz')

        # bipolar colormap
        pos = np.array([0., 0.25, 0.5, 0.75, 1.])
        color = np.array([[0, 0, 0, 255]
                         ,[0, 0, 255, 255]
                         ,[0, 255, 255, 255]
                         ,[255, 255, 0, 255]
                         ,[255, 0, 0, 255]
                         ], dtype=np.ubyte)
        cmap = pg.ColorMap(pos, color)
        lut = cmap.getLookupTable(0.0, 1.0, 256)
        self.img.setLookupTable(lut)

        self.calculate_psd()
        self.render()

    def idx_to_hz(self, idxval):
        conv = self.fs / self.chunksize
        return conv * idxval

    def calcsef(self, perc):
        buf = np.zeros(self.psd.shape[0])
        psdcs = np.cumsum(self.psd, axis=1)
        for i, columncs in enumerate(psdcs):
            thres = perc * columncs[-1] / 100
            sefidx = np.where(columncs > thres)
            try:
                buf[i] = self.idx_to_hz(sefidx[0][0])
            except IndexError:
                pass
        return buf

    def calculate_psd(self):
        nsplit = int(len(self.data) / self.chunksize)
        chunks = self.data[0:nsplit*self.chunksize]
        chunks = np.split(chunks, nsplit)
        chunks = np.vstack(chunks)
        spec = np.fft.rfft(chunks*self.win) / self.chunksize
        self.psd = np.real(spec) ** 2

    def genIndex(self):
        nwindows = self.psd.shape[0]
        winsize = self.chunksize
        windows = np.lib.stride_tricks.as_strided(self.origidx, shape=(nwindows, winsize))
        result = np.apply_along_axis(np.mean, 1, windows)
        index = result.astype(int)
        return index

    def launchSEFExtract(self):
        q = Parameter('SEF percentage', int)
        sefperc = dialogs.askUserValue(q)
        if not sefperc:
            return
        curvename = f'{self.name}-sef{sefperc}'
        sef = self.calcsef(sefperc)
        sefseries = pd.Series(sef, index=self.genIndex(), name=curvename)
        data = {curvename : sefseries}
        plotdata = PlotData(data, name=curvename)
        self.parent.createNewPlotWithData(plotdata)

    def render(self):
        # TODO make lo / hi adjustable
        psdnona = self.psd[~np.isnan(self.psd)]
        hi = np.percentile(psdnona, 95)
        lo = np.percentile(psdnona, 5)
        #print(f'PSD Lo: {lo}, Hi: {hi}')
        self.img.scale(1, self.fs / self.chunksize)
        self.img.setLevels([lo,hi])
        self.img.setImage(self.psd, autoLevels=False)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'poiwidget.ui'
#
# Created by: PyQt5 UI code generator 5.12.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_POISelectorWidget(object):
    def setupUi(self, POISelectorWidget):
        POISelectorWidget.setObjectName("POISelectorWidget")
        POISelectorWidget.resize(763, 601)
        self.horizontalLayout = QtWidgets.QHBoxLayout(POISelectorWidget)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem)
        self.groupBox = QtWidgets.QGroupBox(POISelectorWidget)
        self.groupBox.setCheckable(False)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.radioFixDisabled = QtWidgets.QRadioButton(self.groupBox)
        self.radioFixDisabled.setChecked(True)
        self.radioFixDisabled.setObjectName("radioFixDisabled")
        self.buttonGroup = QtWidgets.QButtonGroup(POISelectorWidget)
        self.buttonGroup.setObjectName("buttonGroup")
        self.buttonGroup.addButton(self.radioFixDisabled)
        self.verticalLayout.addWidget(self.radioFixDisabled)
        self.radioFixMinimum = QtWidgets.QRadioButton(self.groupBox)
        self.radioFixMinimum.setObjectName("radioFixMinimum")
        self.buttonGroup.addButton(self.radioFixMinimum)
        self.verticalLayout.addWidget(self.radioFixMinimum)
        self.radioFixSecondDerivative = QtWidgets.QRadioButton(self.groupBox)
        self.radioFixSecondDerivative.setObjectName("radioFixSecondDerivative")
        self.buttonGroup.addButton(self.radioFixSecondDerivative)
        self.verticalLayout.addWidget(self.radioFixSecondDerivative)
        self.verticalLayout_2.addWidget(self.groupBox)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem1)
        self.horizontalLayout.addLayout(self.verticalLayout_2)

        self.retranslateUi(POISelectorWidget)
        QtCore.QMetaObject.connectSlotsByName(POISelectorWidget)

    def retranslateUi(self, POISelectorWidget):
        _translate = QtCore.QCoreApplication.translate
        POISelectorWidget.setWindowTitle(_translate("POISelectorWidget", "Form"))
        self.groupBox.setTitle(_translate("POISelectorWidget", "Fix Position"))
        self.radioFixDisabled.setText(_translate("POISelectorWidget", "Disabled"))
        self.radioFixMinimum.setText(_translate("POISelectorWidget", "Local Minimum"))
        self.radioFixSecondDerivative.setText(_translate("POISelectorWidget", "2nd derivative peak"))





# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'periodexport.ui'
#
# Created by: PyQt5 UI code generator 5.12.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_PeriodExport(object):
    def setupUi(self, PeriodExport):
        PeriodExport.setObjectName("PeriodExport")
        PeriodExport.resize(255, 237)
        self.verticalLayout = QtWidgets.QVBoxLayout(PeriodExport)
        self.verticalLayout.setObjectName("verticalLayout")
        self.groupBox = QtWidgets.QGroupBox(PeriodExport)
        self.groupBox.setObjectName("groupBox")
        self.formLayout_3 = QtWidgets.QFormLayout(self.groupBox)
        self.formLayout_3.setObjectName("formLayout_3")
        self.btnBrowse = QtWidgets.QPushButton(self.groupBox)
        self.btnBrowse.setObjectName("btnBrowse")
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.btnBrowse)
        self.txtFile = QtWidgets.QLineEdit(self.groupBox)
        self.txtFile.setObjectName("txtFile")
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.txtFile)
        self.verticalLayout.addWidget(self.groupBox)
        self.groupBox_2 = QtWidgets.QGroupBox(PeriodExport)
        self.groupBox_2.setObjectName("groupBox_2")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.groupBox_2)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.lblPeriodStart = QtWidgets.QLabel(self.groupBox_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lblPeriodStart.sizePolicy().hasHeightForWidth())
        self.lblPeriodStart.setSizePolicy(sizePolicy)
        self.lblPeriodStart.setAlignment(QtCore.Qt.AlignCenter)
        self.lblPeriodStart.setObjectName("lblPeriodStart")
        self.horizontalLayout_2.addWidget(self.lblPeriodStart)
        self.label_3 = QtWidgets.QLabel(self.groupBox_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_3.sizePolicy().hasHeightForWidth())
        self.label_3.setSizePolicy(sizePolicy)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_2.addWidget(self.label_3)
        self.lblPeriodStop = QtWidgets.QLabel(self.groupBox_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lblPeriodStop.sizePolicy().hasHeightForWidth())
        self.lblPeriodStop.setSizePolicy(sizePolicy)
        self.lblPeriodStop.setAlignment(QtCore.Qt.AlignCenter)
        self.lblPeriodStop.setObjectName("lblPeriodStop")
        self.horizontalLayout_2.addWidget(self.lblPeriodStop)
        self.verticalLayout.addWidget(self.groupBox_2)
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.label = QtWidgets.QLabel(PeriodExport)
        self.label.setObjectName("label")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label)
        self.txtPatient = QtWidgets.QLineEdit(PeriodExport)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.txtPatient.sizePolicy().hasHeightForWidth())
        self.txtPatient.setSizePolicy(sizePolicy)
        self.txtPatient.setObjectName("txtPatient")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.txtPatient)
        self.label_7 = QtWidgets.QLabel(PeriodExport)
        self.label_7.setObjectName("label_7")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_7)
        self.txtPeriod = QtWidgets.QComboBox(PeriodExport)
        self.txtPeriod.setEditable(True)
        self.txtPeriod.setObjectName("txtPeriod")
        self.txtPeriod.addItem("")
        self.txtPeriod.addItem("")
        self.txtPeriod.addItem("")
        self.txtPeriod.addItem("")
        self.txtPeriod.addItem("")
        self.txtPeriod.addItem("")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.txtPeriod)
        self.label_2 = QtWidgets.QLabel(PeriodExport)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.txtComment = QtWidgets.QLineEdit(PeriodExport)
        self.txtComment.setObjectName("txtComment")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.txtComment)
        self.verticalLayout.addLayout(self.formLayout)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.btnOk = QtWidgets.QPushButton(PeriodExport)
        self.btnOk.setObjectName("btnOk")
        self.horizontalLayout_3.addWidget(self.btnOk)
        self.btnCancel = QtWidgets.QPushButton(PeriodExport)
        self.btnCancel.setObjectName("btnCancel")
        self.horizontalLayout_3.addWidget(self.btnCancel)
        self.verticalLayout.addLayout(self.horizontalLayout_3)

        self.retranslateUi(PeriodExport)
        QtCore.QMetaObject.connectSlotsByName(PeriodExport)

    def retranslateUi(self, PeriodExport):
        _translate = QtCore.QCoreApplication.translate
        PeriodExport.setWindowTitle(_translate("PeriodExport", "Export period information"))
        self.groupBox.setTitle(_translate("PeriodExport", "Destination file"))
        self.btnBrowse.setText(_translate("PeriodExport", "Browse"))
        self.groupBox_2.setTitle(_translate("PeriodExport", "Period"))
        self.lblPeriodStart.setText(_translate("PeriodExport", "Beginning"))
        self.label_3.setText(_translate("PeriodExport", "-"))
        self.lblPeriodStop.setText(_translate("PeriodExport", "End"))
        self.label.setText(_translate("PeriodExport", "Patient"))
        self.label_7.setText(_translate("PeriodExport", "Period Identification"))
        self.txtPeriod.setItemText(0, _translate("PeriodExport", "carotid"))
        self.txtPeriod.setItemText(1, _translate("PeriodExport", "ascending"))
        self.txtPeriod.setItemText(2, _translate("PeriodExport", "descending"))
        self.txtPeriod.setItemText(3, _translate("PeriodExport", "renal"))
        self.txtPeriod.setItemText(4, _translate("PeriodExport", "iliac"))
        self.txtPeriod.setItemText(5, _translate("PeriodExport", "bolus"))
        self.label_2.setText(_translate("PeriodExport", "Comment"))
        self.btnOk.setText(_translate("PeriodExport", "Ok"))
        self.btnCancel.setText(_translate("PeriodExport", "Cancel"))





import os, csv
from itertools import zip_longest
from functools import partial

import pandas as pd

from graphysio.dialogs import DlgPeriodExport, askDirPath, askFilePath
from graphysio.utils import sanitize_filename


class TsExporter():
    periodfields = ['patient', 'begin', 'end', 'periodid', 'comment']

    def __init__(self, parent, name):
        self.parent = parent
        self.name = name
        try:
            self.outdir = parent.properties['dircache']
        except KeyError:
            self.outdir = os.path.expanduser('~')

    def seriestocsv(self) -> None:
        filepath = askFilePath('Export to', f'{self.name}.csv', self.outdir)
        if filepath is None:
            return
        self.outdir = os.path.dirname(filepath)
        xmin, xmax = self.parent.vbrange
        series = [curve.series[~curve.series.index.duplicated()] for curve in self.parent.curves.values()]
        data = pd.concat(series, axis=1).sort_index()
        data['datetime'] = pd.to_datetime(data.index, unit = 'ns')
        data = data.loc[xmin:xmax]
        data.to_csv(filepath, date_format="%Y-%m-%d %H:%M:%S.%f", index_label='timens')

    def periodstocsv(self):
        xmin, xmax = self.parent.vbrange
        dlg = DlgPeriodExport(begin   = xmin,
                              end     = xmax,
                              patient = self.name,
                              directory = self.outdir)
        if not dlg.exec_():
            return
        self.name = dlg.patient
        self.outdir = os.path.dirname(dlg.filepath)

        with open(dlg.filepath, 'a', newline='') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=self.periodfields, quoting=csv.QUOTE_MINIMAL)
            if not os.path.exists(dlg.filepath):
                writer.writeheader()
            writer.writerow({'patient'  : dlg.patient,
                             'begin'    : xmin,
                             'end'      : xmax,
                             'periodid' : dlg.periodname,
                             'comment'  : dlg.comment})

    def cyclestocsv(self) -> None:
        outdir = askDirPath("Export to", self.outdir)
        if outdir is None:
            # Cancel pressed
            return

        # Some non trivial manipulations to get the cycles from all
        # curves, then reorganize to group by the n-th cycle from each
        # curve and put those cycles into a dataframe for export.
        def getCurveCycles(curve):
            cycleIdx = curve.getCycleIndices()
            cycles = (curve.series.loc[b:b+d] for b, d in zip(*cycleIdx))
            return cycles

        curves = self.parent.curves.values()
        allByCurve = (getCurveCycles(curve) for curve in curves)
        allByCycle = zip_longest(*allByCurve)

        for n, cycle in enumerate(allByCycle):
            idxstart = None
            for s in cycle:
                # Make all series start at the same point
                if s is None or len(s) < 1:
                    continue
                if idxstart is None:
                    idxstart = s.index[0]
                idxdelta = s.index[0] - idxstart
                s.index -= idxdelta
            df = pd.concat(cycle, axis=1)
            df['datetime'] = pd.to_datetime(df.index, unit = 'ns')

            filename = sanitize_filename(f'{self.name}-{n+1}.csv')
            filepath = os.path.join(self.outdir, filename)
            df.to_csv(filepath, date_format="%Y-%m-%d %H:%M:%S.%f", index_label='timens')

    def cyclepointstocsv(self) -> None:
        filepath = askFilePath('Export to', f'{self.name}-feet.csv', self.outdir)
        if filepath is None:
            # Cancel pressed
            return
        self.outdir = os.path.dirname(filepath)

        feetseries = []
        for c in self.parent.curves.values():
            for k, v in c.feetitem.indices.items():
                feetseries.append(pd.Series(v, name=f'{c.name()}-{k}'))
        df = pd.concat(feetseries, axis=1)
        df.to_csv(filepath, index_label='idx')

class PuExporter():
    def __init__(self, parent, name):
        self.parent = parent
        self.name = name
        try:
            self.outdir = parent.properties['dircache']
        except KeyError:
            self.outdir = os.path.expanduser('~')

    def exportloops(self):
        outdir = askDirPath('Export to', self.outdir)
        if outdir is None:
            # Cancel pressed
            return
        self.outdir = outdir
        self.writetable()
        self.writeloops()

    def writetable(self):
        data = []
        for loop in self.parent.loops:
            alpha, beta, gala = loop.angles
            delay = loop.offset / 1e6
            tmpdict = {'alpha' : alpha, 'beta' : beta, 'gala' : gala, 'delay' : delay}
            data.append(tmpdict)
        df = pd.DataFrame(data)
        df.index += 1
        filename = sanitize_filename(f'{self.name}-loopdata.csv')
        filepath = os.path.join(self.outdir, filename)
        df.to_csv(filepath, index_label='idx')

    def writeloops(self):
        for n, loop in enumerate(self.parent.loops):
            df = loop.df
            filename = sanitize_filename(f'{self.name}-{n+1}.csv')
            filepath = os.path.join(self.outdir, filename)
            df['datetime'] = pd.to_datetime(df.index, unit='ns')
            df.to_csv(filepath, date_format="%Y-%m-%d %H:%M:%S.%f", index_label='timens')

class POIExporter():
    def __init__(self, parent, name):
        self.parent = parent
        self.name = name
        try:
            self.outdir = parent.properties['dircache']
        except KeyError:
            self.outdir = os.path.expanduser('~')

    def poitocsv(self):
        filepath = askFilePath('Export to', f'{self.name}-poi.csv', self.outdir)
        if filepath is None:
            # Cancel pressed
            return
        self.outdir = os.path.dirname(filepath)

        srcseries = self.parent.curve.series
        poiidx = self.parent.curve.feetitem.indices[self.parent.pointkey]
        pois = srcseries[poiidx.dropna()].rename('poi')
        sers = [srcseries, pois]
        df = pd.concat(sers, axis=1, keys=[s.name for s in sers])
        df['datetime'] = pd.to_datetime(df.index, unit = 'ns')
        df.to_csv(filepath, date_format="%Y-%m-%d %H:%M:%S.%f", index_label='timens')



import os
import csv
from functools import partial

import numpy as np
import pandas as pd

from pyqtgraph.Qt import QtGui, QtCore, QtWidgets

from graphysio import types, ui


class CsvReader(QtCore.QRunnable):
    def __init__(self, csvrequest, sigdata, sigerror):
        super().__init__()
        self.csvrequest = csvrequest
        self.sigdata = sigdata
        self.sigerror = sigerror

    def run(self) -> None:
        try:
            data = self.getdata()
        except Exception as e:
            self.sigerror.emit(e)
        else:
            self.sigdata.emit(data)

    def getdata(self) -> types.PlotData:
        data = pd.read_csv(self.csvrequest.filepath,
                           sep       = self.csvrequest.seperator,
                           usecols   = self.csvrequest.fields,
                           decimal   = self.csvrequest.decimal,
                           skiprows  = self.csvrequest.droplines,
                           encoding  = self.csvrequest.encoding,
                           index_col = False,
                           engine    = 'c')

        pdtonum = partial(pd.to_numeric, errors='coerce')
        dtformat = self.csvrequest.datetime_format
        if self.csvrequest.generatex:
            data.index = (1e9 * data.index / self.csvrequest.samplerate).astype(np.int64)
            # Make all data numeric and remove empty rows
            data = data.apply(pdtonum)
            data = data.dropna(axis='rows', how='all')
        else:
            timestamp = data[self.csvrequest.dtfield]
            data = data.drop(columns=self.csvrequest.dtfield)
            # Force all columns to numeric
            data = data.apply(pdtonum)

            if dtformat == '<seconds>':
                timestamp = pdtonum(timestamp)
                timestamp = pd.to_datetime(timestamp * 1e9, unit='ns')
            elif dtformat == '<milliseconds>':
                timestamp = pdtonum(timestamp)
                timestamp = pd.to_datetime(timestamp * 1e6, unit='ns')
            elif dtformat == '<microseconds>':
                timestamp = pdtonum(timestamp)
                timestamp = pd.to_datetime(timestamp * 1e3, unit='ns')
            elif dtformat == '<nanoseconds>':
                timestamp = pdtonum(timestamp)
                timestamp = pd.to_datetime(timestamp, unit='ns')
            elif dtformat == '<infer>':
                timestamp = pd.to_datetime(timestamp, infer_datetime_format=True)
            else:
                timestamp = pd.to_datetime(timestamp, format=dtformat)

            # Convert timestamp to UTC and use as index
            timestamp = pd.Index(timestamp).tz_localize(self.csvrequest.timezone).tz_convert('UTC')
            timestamp = timestamp.astype(np.int64)
            data = data.set_index([timestamp])

        data = data.dropna(axis='columns', how='all')
        data = data.sort_index()

        plotdata = types.PlotData(data=data, filepath=self.csvrequest.filepath)
        return plotdata


class DlgNewPlotCsv(ui.Ui_NewPlot, QtWidgets.QDialog):
    dlgdata = QtCore.pyqtSignal(object)

    def __init__(self, parent=None, title="New Plot", directory=""):
        super().__init__(parent=parent)
        self.setupUi(self)
        self.setWindowTitle(title)

        self.dircache = directory
        self.csvrequest = types.CsvRequest()

        # Attach models to ListViews
        self.lstX = QtGui.QStandardItemModel()
        self.lstY = QtGui.QStandardItemModel()
        self.lstAll = QtGui.QStandardItemModel()

        self.lstVX.setModel(self.lstX)
        self.lstVY.setModel(self.lstY)
        self.lstVAll.setModel(self.lstAll)

        # Setup Field Table
        self.lstVAll.verticalHeader().hide()
        self.lstVAll.horizontalHeader().setResizeMode(QtGui.QHeaderView.Stretch)
        self.lstVAll.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)

        # Connect callbacks
        self.btnBrowse.clicked.connect(self.selectFile)
        self.btnLoad.clicked.connect(self.loadCsvFields)
        self.btnOk.clicked.connect(self.loadPlot)
        self.btnCancel.clicked.connect(self.reject)
        self.btnToX.clicked.connect(self.moveToX)
        self.btnToY.clicked.connect(self.moveToY)
        self.btnRemoveX.clicked.connect(self.delFromX)
        self.btnRemoveY.clicked.connect(self.delFromY)
        self.lstVX.currentIndexChanged.connect(self.xChanged)

    # Methods / Callbacks
    def selectFile(self):
        filepath = QtGui.QFileDialog.getOpenFileName(parent = self,
                                                     caption = "Open CSV file",
                                                     directory = self.dircache)
        # PyQt5 API change
        if type(filepath) is not str:
            filepath = filepath[0]

        if not filepath:
            return
        self.dircache = os.path.dirname(filepath)
        self.txtFile.setText(filepath)
        # Guesstimate CSV field and decimal seperators
        delims = self.estimateDelimiters(filepath)
        self.txtSep.setEditText(delims[0])
        self.txtDecimal.setEditText(delims[1])
        self.txtDateTime.setEditText(f'%Y-%m-%d %H:%M:%S{delims[1]}%f')

    def estimateDelimiters(self, filepath):
        encoding = self.txtEncoding.currentText()
        with open(filepath, 'r', encoding=encoding) as csvfile:
            line1 = next(csvfile)
            line2 = next(csvfile)
            semipos = line1.find(';')
            if semipos == -1:
                seperator = ','
            else:
                seperator = ';'
            periodpos = line2.find('.')
            if periodpos == -1:
                decimal = ','
            else:
                decimal = '.'
        return (seperator, decimal)

    def loadCsvFields(self):
        sep = self.txtSep.currentText()
        if sep == '<tab>':
            sep = '\t'
        filepath = self.txtFile.text()
        # Use the csv module to retrieve csv fields
        for lst in [self.lstAll, self.lstX, self.lstY]:
            lst.clear()
        self.lstAll.setHorizontalHeaderLabels(["Field", "1st Line"])
        encoding = self.txtEncoding.currentText()
        with open(filepath, 'r', encoding=encoding) as csvfile:
            # Artificially drop n first lines as requested
            for _ in range(self.spnLinedrop.value()):
                next(csvfile)
            csvreader = csv.DictReader(csvfile, delimiter=sep)
            row = next(csvreader)
            for key, value in row.items():
                if key is None:
                    continue
                keyitem = QtGui.QStandardItem(key)
                valueitem = QtGui.QStandardItem(value)
                self.lstAll.appendRow([keyitem, valueitem])
        self.lstAll.sort(0)

    def xChanged(self, newtext):
        if self.lstX.rowCount() > 0:
            self.chkGenX.setCheckState(QtCore.Qt.Unchecked)
        else:
            self.chkGenX.setCheckState(QtCore.Qt.Checked)

    def moveToX(self):
        if self.lstX.rowCount() > 0:
            # Only allow one element for X.
            return
        selection = self.lstVAll.selectedIndexes()
        rowindex = selection[0].row()
        row = self.lstAll.takeRow(rowindex)
        self.lstX.appendRow(row)

    def moveToY(self):
        while True:
            selection = self.lstVAll.selectedIndexes()
            if len(selection) < 1:
                break
            rowindex = selection[0].row()
            self.lstY.appendRow(self.lstAll.takeRow(rowindex))

    def delFromX(self):
        try:
            row = self.lstX.takeRow(0)
        except IndexError:
            return
        self.lstAll.appendRow(row)

    def delFromY(self):
        while True:
            rowindexes = self.lstVY.selectedIndexes()
            if len(rowindexes) < 1:
                break
            row = rowindexes[0].row()
            self.lstAll.appendRow(self.lstY.takeRow(row))

    def loadPlot(self):
        yRows = [i.text() for i in self.lstY.findItems("", QtCore.Qt.MatchContains)]
        xRows = [i.text() for i in self.lstX.findItems("", QtCore.Qt.MatchContains)]

        seperator = self.txtSep.currentText()
        if seperator == '<tab>':
            self.csvrequest.seperator = '\t'
        else:
            self.csvrequest.seperator = seperator

        self.csvrequest.generatex = (self.chkGenX.checkState() > QtCore.Qt.Unchecked)
        if self.csvrequest.generatex or len(xRows) < 1:
            self.csvrequest.dtfield = None
        else:
            self.csvrequest.dtfield = xRows[0]

        self.csvrequest.samplerate = self.spnFs.value()
        self.csvrequest.yfields = yRows
        self.csvrequest.filepath = self.txtFile.text()
        self.csvrequest.decimal = self.txtDecimal.currentText()
        self.csvrequest.datetime_format = self.txtDateTime.currentText()
        self.csvrequest.droplines = self.spnLinedrop.value()
        self.csvrequest.encoding = self.txtEncoding.currentText()
        self.csvrequest.timezone = self.txtTimezone.currentText()

        self.dlgdata.emit(self.csvrequest)
        self.accept()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'curveproperties.ui'
#
# Created by: PyQt5 UI code generator 5.12.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_CurveProperties(object):
    def setupUi(self, CurveProperties):
        CurveProperties.setObjectName("CurveProperties")
        CurveProperties.resize(342, 280)
        self.verticalLayout = QtWidgets.QVBoxLayout(CurveProperties)
        self.verticalLayout.setObjectName("verticalLayout")
        self.grpName = QtWidgets.QGroupBox(CurveProperties)
        self.grpName.setObjectName("grpName")
        self.formLayout = QtWidgets.QFormLayout(self.grpName)
        self.formLayout.setObjectName("formLayout")
        self.txtName = QtWidgets.QLineEdit(self.grpName)
        self.txtName.setObjectName("txtName")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.txtName)
        self.label_3 = QtWidgets.QLabel(self.grpName)
        self.label_3.setObjectName("label_3")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.LabelRole, self.label_3)
        self.btnColor = QtWidgets.QPushButton(self.grpName)
        self.btnColor.setObjectName("btnColor")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.FieldRole, self.btnColor)
        self.label_4 = QtWidgets.QLabel(self.grpName)
        self.label_4.setObjectName("label_4")
        self.formLayout.setWidget(7, QtWidgets.QFormLayout.LabelRole, self.label_4)
        self.lblSamplerate = QtWidgets.QLabel(self.grpName)
        self.lblSamplerate.setObjectName("lblSamplerate")
        self.formLayout.setWidget(7, QtWidgets.QFormLayout.FieldRole, self.lblSamplerate)
        self.cmbSymbol = QtWidgets.QComboBox(self.grpName)
        self.cmbSymbol.setObjectName("cmbSymbol")
        self.cmbSymbol.addItem("")
        self.cmbSymbol.addItem("")
        self.cmbSymbol.addItem("")
        self.cmbSymbol.addItem("")
        self.cmbSymbol.addItem("")
        self.cmbSymbol.addItem("")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.cmbSymbol)
        self.label_6 = QtWidgets.QLabel(self.grpName)
        self.label_6.setObjectName("label_6")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.label_6)
        self.label_2 = QtWidgets.QLabel(self.grpName)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.spnWidth = QtWidgets.QSpinBox(self.grpName)
        self.spnWidth.setObjectName("spnWidth")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.FieldRole, self.spnWidth)
        self.label_5 = QtWidgets.QLabel(self.grpName)
        self.label_5.setObjectName("label_5")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_5)
        self.label = QtWidgets.QLabel(self.grpName)
        self.label.setObjectName("label")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label)
        self.cmbConnect = QtWidgets.QComboBox(self.grpName)
        self.cmbConnect.setObjectName("cmbConnect")
        self.cmbConnect.addItem("")
        self.cmbConnect.addItem("")
        self.cmbConnect.addItem("")
        self.cmbConnect.addItem("")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.cmbConnect)
        self.verticalLayout.addWidget(self.grpName)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.okButton = QtWidgets.QPushButton(CurveProperties)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.okButton.sizePolicy().hasHeightForWidth())
        self.okButton.setSizePolicy(sizePolicy)
        self.okButton.setObjectName("okButton")
        self.horizontalLayout.addWidget(self.okButton)
        self.cancelButton = QtWidgets.QPushButton(CurveProperties)
        self.cancelButton.setObjectName("cancelButton")
        self.horizontalLayout.addWidget(self.cancelButton)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(CurveProperties)
        QtCore.QMetaObject.connectSlotsByName(CurveProperties)

    def retranslateUi(self, CurveProperties):
        _translate = QtCore.QCoreApplication.translate
        CurveProperties.setWindowTitle(_translate("CurveProperties", "Dialog"))
        self.grpName.setTitle(_translate("CurveProperties", "GroupBox"))
        self.label_3.setText(_translate("CurveProperties", "Color"))
        self.btnColor.setText(_translate("CurveProperties", "Change Color"))
        self.label_4.setText(_translate("CurveProperties", "Sampling rate"))
        self.lblSamplerate.setText(_translate("CurveProperties", "TextLabel"))
        self.cmbSymbol.setItemText(0, _translate("CurveProperties", "None"))
        self.cmbSymbol.setItemText(1, _translate("CurveProperties", "o"))
        self.cmbSymbol.setItemText(2, _translate("CurveProperties", "s"))
        self.cmbSymbol.setItemText(3, _translate("CurveProperties", "t"))
        self.cmbSymbol.setItemText(4, _translate("CurveProperties", "d"))
        self.cmbSymbol.setItemText(5, _translate("CurveProperties", "+"))
        self.label_6.setText(_translate("CurveProperties", "Connect Points"))
        self.label_2.setText(_translate("CurveProperties", "Line width"))
        self.label_5.setText(_translate("CurveProperties", "Symbol"))
        self.label.setText(_translate("CurveProperties", "Name"))
        self.cmbConnect.setItemText(0, _translate("CurveProperties", "None"))
        self.cmbConnect.setItemText(1, _translate("CurveProperties", "All"))
        self.cmbConnect.setItemText(2, _translate("CurveProperties", "Pairs"))
        self.cmbConnect.setItemText(3, _translate("CurveProperties", "Finite"))
        self.okButton.setText(_translate("CurveProperties", "Ok"))
        self.cancelButton.setText(_translate("CurveProperties", "Cancel"))





from typing import Optional

import numpy as np
import pandas as pd
import pyqtgraph as pg

from pyqtgraph.Qt import QtGui, QtCore

from graphysio.legend import LegendItem
from graphysio.utils import Colors
from graphysio.plotwidgets import curves


class TimeAxisItem(pg.AxisItem):
    def tickStrings(self, values, scale, spacing):
        ret = []
        for i, value in enumerate(values):
            value = value / 1e6 # convert from ns to ms
            date = QtCore.QDateTime.fromMSecsSinceEpoch(value)
            date = date.toTimeSpec(QtCore.Qt.UTC)
            datestr = date.toString("dd/MM/yyyy\nhh:mm:ss.zzz")
            ret.append(datestr)
        return ret


class PlotWidget(pg.PlotWidget):
    def __init__(self, name, parent=None):
        self.parent = parent
        self.name = name
        self.colors = Colors()
        self.hiddenitems = []
        self.properties = {}

        axisItems = {'bottom': TimeAxisItem(orientation='bottom')}
        super().__init__(parent=parent, axisItems=axisItems, background='w')

        self.legend = LegendItem()
        self.legend.setParentItem(self.getPlotItem())

        self.vb = self.getViewBox()
        self.vb.setMouseMode(self.vb.RectMode)

        self.setCursor(QtCore.Qt.CrossCursor)

    def appendData(self, newplotdata, dorealign=False):
        for seriesname in newplotdata.data:
            self.addSeriesAsCurve(newplotdata.data[seriesname], dorealign=dorealign)

    @property
    def curves(self):
        return {item.name() : item for item in self.listDataItems() if isinstance(item, curves.CurveItem)}

    def getPen(self):
        return next(self.colors)

    def addSeriesAsCurve(self, series, pen=None, dorealign=False, withfeet=True):
        if len(series) < 1:
            return
        if dorealign and self.curves:
            # Timeshift new curves to make the beginnings coincide
            begins = [curve.series.index[0] for curve in self.curves.values()]
            offset = min(begins) - series.index[0]
            series.index += offset
        try:
            # Append to existing curve?
            curve = self.curves[series.name]
            curve.extend(series)
        except KeyError:
            # New curve
            if pen is None:
                pen = self.getPen()
            if withfeet:
                Curve = curves.CurveItemWithPOI
            else:
                Curve = curves.CurveItem
            curve = Curve(series=series, parent=self, pen=pen)
            self.addCurve(curve)
        return curve

    def addCurve(self, curve, pen=None):
        if pen is not None:
            curve.setPen(pen)
        self.addItem(curve)
        if isinstance(curve, pg.PlotDataItem):
            self.legend.addItem(curve, curve.name())
        curve.visible.emit()

    def removeCurve(self, curve):
        self.removeItem(curve)
        self.legend.removeItem(curve.name())
        curve.invisible.emit()

    def validateNewCurveName(self, proposedname: str, alwaysShow: bool = False) -> Optional[str]:
        if proposedname not in self.curves and not alwaysShow:
            return proposedname
        newname, okPressed = QtGui.QInputDialog.getText(self, "Series name", "Series with identical names will be merged.", text=proposedname)
        if not okPressed:
            return None
        return newname

    @property
    def vbrange(self):
        vbrange = self.vb.viewRange()
        xmin, xmax = map(int, vbrange[0])
        return (xmin, xmax)

    @property
    def menu(self):
        return {}



#!/usr/bin/env python3

import sys
from pyqtgraph.Qt import QtGui
from graphysio.mainui import MainUi

import traceback
import logging, logging.handlers

if __name__ == '__main__':
    app = QtGui.QApplication(sys.argv)

    winmain = MainUi()
    winmain.show()

    log = logging.getLogger("GraPhysio")
    smtph = logging.handlers.SMTPHandler('puff.joachim.cc', 'graphysio@example.com', 'jaj_graphysio@joachim.cc', 'GraPhysio_error')
    log.addHandler(smtph)

    status = 1
    try:
        status = app.exec_()
    except:
        tb = traceback.format_exc()
        log.error(tb)
    finally:
        sys.exit(status)



from .csv import CsvReader, DlgNewPlotCsv
from .parquet import ParquetReader, DlgNewPlotParquet

readers = {'csv' : CsvReader, 'parquet' : ParquetReader}
dlgNewPlots = {'csv' : DlgNewPlotCsv, 'parquet' : DlgNewPlotParquet}



from typing import Optional
from datetime import datetime
import pathlib

from pint import UnitRegistry
from pint.errors import (DimensionalityError, UndefinedUnitError)

from pyqtgraph.Qt import QtGui, QtCore, QtWidgets
import pyqtgraph as pg

from graphysio import types, ui
from graphysio.algorithms import filters
from graphysio.utils import sanitize_filepath

ureg = UnitRegistry()


class DlgCycleDetection(ui.Ui_CycleDetection, QtWidgets.QDialog):
    dlgdata = QtCore.pyqtSignal(object)

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setupUi(self)

        self.okButton.clicked.connect(self.accept)
        self.cancelButton.clicked.connect(self.reject)

        self.choices = {}

        plotframe = self.parent().tabWidget.currentWidget()
        if plotframe is None:
            return

        for n, curvename in enumerate(plotframe.curves.keys()):
            combo = QtGui.QComboBox()
            combo.addItems([ft.value for ft in types.CycleId])
            curveitem = QtGui.QTableWidgetItem(curvename)

            self.table.insertRow(n)
            self.table.setItem(n, 0, curveitem)
            self.table.setCellWidget(n, 1, combo)
            self.choices[curvename] = combo

        self.table.horizontalHeader().setResizeMode(QtGui.QHeaderView.ResizeToContents)

    def accept(self):
        result = {curve: combo.currentText() for (curve, combo) in self.choices.items()}
        self.dlgdata.emit(result)
        super().accept()


class DlgFilter(ui.Ui_Filter, QtWidgets.QDialog):
    dlgdata = QtCore.pyqtSignal(object)

    def __init__(self, parent=None, filterfeet=False):
        super().__init__(parent=parent)
        self.setupUi(self)

        self.okButton.clicked.connect(self.accept)
        self.cancelButton.clicked.connect(self.reject)

        self.choices = {}

        plotframe = self.parent().tabWidget.currentWidget()
        if not plotframe:
            return

        def fillTable(items, itemtype, filters):
            rowoffset = self.table.rowCount()
            for n, itemname in enumerate(items):
                combo = QtGui.QComboBox()

                combo.addItems(['None'])
                filternames = list(filters.keys())
                combo.addItems(filternames)

                curveitem = QtGui.QTableWidgetItem(itemname)

                idx = rowoffset + n
                self.table.insertRow(idx)
                self.table.setItem(idx, 0, curveitem)
                self.table.setCellWidget(idx, 1, combo)
                self.choices[itemname] = (combo, itemtype)

        curves = list(plotframe.curves.keys())
        if filterfeet:
            self.chkNewcurve.hide()
            fillTable(curves, 'curve', filters.FeetFilters)
        else:
            fillTable(curves, 'curve', filters.Filters)

        self.table.horizontalHeader().setResizeMode(QtGui.QHeaderView.ResizeToContents)

    def accept(self):
        curvefilters = {}
        for itemname, value in self.choices.items():
            combo, itemtype = value
            curvefilters[itemname] = combo.currentText()
        createnew = (self.chkNewcurve.checkState() > QtCore.Qt.Unchecked)
        result = (createnew, curvefilters)
        self.dlgdata.emit(result)
        super().accept()


class DlgSetupPULoop(ui.Ui_SetupPULoop, QtWidgets.QDialog):
    dlgdata = QtCore.pyqtSignal(object)

    def __init__(self, sourcewidget, parent=None):
        super().__init__(parent=parent)
        self.setupUi(self)

        self.okButton.clicked.connect(self.accept)
        self.cancelButton.clicked.connect(self.reject)

        try:
            curvenames = list(sourcewidget.curves.keys())
        except AttributeError:
            return

        self.comboU.addItems(curvenames)
        self.comboP.addItems(curvenames)

    def accept(self):
        uname = self.comboU.currentText()
        pname = self.comboP.currentText()
        result = (uname, pname)
        self.dlgdata.emit(result)
        super().accept()


class DlgPeriodExport(ui.Ui_PeriodExport, QtWidgets.QDialog):
    def __init__(self, begin, end, patient="", directory="", parent=None):
        super().__init__(parent=parent)
        self.setupUi(self)

        self.dircache = directory

        self.lblPeriodStart.setText(str(begin))
        self.lblPeriodStop.setText(str(end))
        self.txtPatient.setText(patient)

        self.btnOk.clicked.connect(self.accept)
        self.btnCancel.clicked.connect(self.reject)
        self.btnBrowse.clicked.connect(self.selectFile)

    def selectFile(self):
        filepath = QtGui.QFileDialog.getSaveFileName(caption = "Export to",
                                                     filter  = "CSV files (*.csv *.dat)",
                                                     options = QtGui.QFileDialog.DontConfirmOverwrite,
                                                     directory = self.dircache)
        # PyQt5 API change
        if type(filepath) is not str:
            filepath = filepath[0]

        if filepath:
            self.txtFile.setText(filepath)
            self.dircache = os.path.dirname(filepath)

    @property
    def patient(self):
        return self.txtPatient.text()

    @property
    def comment(self):
        return self.txtComment.text()

    @property
    def periodname(self):
        return self.txtPeriod.currentText()

    @property
    def filepath(self):
        return self.txtFile.text()

class DlgCurveSelection(ui.Ui_CurveSelection, QtWidgets.QDialog):
    def __init__(self, visible=[], hidden=[], parent=None):
        super().__init__(parent=parent)
        self.setupUi(self)

        self.okButton.clicked.connect(self.accept)
        self.cancelButton.clicked.connect(self.reject)
        self.btnProperties.clicked.connect(self.openProperties)

        hiddenhash = {curve.name(): curve for curve in hidden}
        self.curvehash = {curve.name(): curve for curve in visible}
        self.curvehash.update(hiddenhash)

        for curve in visible:
            self.addCurve(curve.name(), checked=True)
        for curve in hidden:
            self.addCurve(curve.name(), checked=False)

    def openProperties(self):
        selected = self.lstCurves.selectedItems()
        try:
            curvename = selected[0].text()
        except IndexError:
            return
        curve = self.curvehash[curvename]
        dlg = DlgCurveProperties(curve)
        dlg.exec_()

    def addCurve(self, name, checked):
        item = QtGui.QListWidgetItem()
        item.setText(name)
        item.setFlags(item.flags() | QtCore.Qt.ItemIsUserCheckable)
        if checked:
            item.setCheckState(QtCore.Qt.Checked)
        else:
            item.setCheckState(QtCore.Qt.Unchecked)
        self.lstCurves.addItem(item)

    @property
    def result(self):
        items = self.lstCurves.findItems("", QtCore.Qt.MatchContains)
        ischecked = lambda item: not (item.checkState() == QtCore.Qt.Unchecked)
        checked = list(filter(ischecked, items))
        unchecked = [item for item in items if item not in checked]
        visible = [self.curvehash[item.text()] for item in checked]
        invisible = [self.curvehash[item.text()] for item in unchecked]
        return (visible, invisible)

class DlgCurveProperties(ui.Ui_CurveProperties, QtWidgets.QDialog):
    def __init__(self, curve, parent=None):
        super().__init__(parent=parent)
        self.setupUi(self)

        self.curve = curve

        self.okButton.clicked.connect(self.ok)
        self.cancelButton.clicked.connect(self.reject)
        self.btnColor.clicked.connect(self.chooseColor)

        connect = curve.opts['connect']
        symbol = curve.opts['symbol']
        if symbol is None:
            symbol = 'None'

        pen = curve.opts['pen']
        if isinstance(pen, QtGui.QPen):
            color = pen.color()
            width = pen.width()
        else:
            color = pen
            width = 1
        self.color = color

        idx = self.cmbConnect.findText(connect, QtCore.Qt.MatchFixedString)
        self.cmbConnect.setCurrentIndex(idx)

        idx = self.cmbSymbol.findText(symbol, QtCore.Qt.MatchFixedString)
        self.cmbSymbol.setCurrentIndex(idx)

        self.cmbSymbol.setEditText(symbol)
        self.grpName.setTitle(curve.name())
        self.txtName.setText(curve.name())
        self.spnWidth.setValue(width)
        self.btnColor.setStyleSheet(f'background-color: {color.name()}')
        self.lblSamplerate.setText(str(curve.samplerate))

    def ok(self):
        connect = str(self.cmbConnect.currentText().lower())
        if connect == 'none':
            connect = [0]
        symbol = self.cmbSymbol.currentText().lower()
        if symbol == 'none':
            symbol = None

        #self.curve.setData(connect=connect, symbol=symbol)
        self.curve.setData(symbol=symbol)

        width = self.spnWidth.value()
        pen = pg.mkPen(color=self.color, width=width)
        self.curve.setPen(pen)

        self.accept()

    def chooseColor(self):
        color = QtGui.QColorDialog.getColor()
        if not color.isValid():
            return
        self.color = color
        self.btnColor.setStyleSheet(f'background-color: {color.name()}')

class DlgSetDateTime(ui.Ui_SetDateTime, QtWidgets.QDialog):
    def __init__(self, parent=None, prevdatetime=None):
        super().__init__(parent=parent)
        self.setupUi(self)

        self.timestamp = 0

        self.btnOk.clicked.connect(self.ok)
        self.btnCancel.clicked.connect(self.reject)

        if prevdatetime is None:
            datetime = QtCore.QDateTime.currentDateTime()
        else:
            datetime = QtCore.QDateTime.fromMSecsSinceEpoch(prevdatetime / 1e6)
        curdate = datetime.date()
        curtime = datetime.time()
        self.widgetCalendar.setSelectedDate(curdate)
        self.edDate.setDate(curdate)
        self.edTime.setTime(curtime)

    def ok(self):
        date = self.edDate.date()
        time = self.edTime.time()
        dt = QtCore.QDateTime(date, time, QtCore.Qt.UTC)
        self.timestamp = dt.toMSecsSinceEpoch() * 1e6
        self.accept()

    @property
    def result(self):
        return self.timestamp

class DlgCurveAlgebra(QtWidgets.QDialog):
    def __init__(self, parent=None, curvecorr={}):
        super().__init__(parent=parent)
        self.setupUi(curvecorr)
    
    def setupUi(self, curvecorr):
        vstack = QtWidgets.QVBoxLayout(self)

        self.lbl = QtWidgets.QLabel('Enter formula for new curve:')
        vstack.addWidget(self.lbl)

        self.formula = QtWidgets.QLineEdit('a^2 + log(2*b)')
        vstack.addWidget(self.formula)

        curveslbl = '\n'.join([f'{x}: {y}' for x,y in curvecorr.items()])
        self.curveletters = QtWidgets.QLabel(curveslbl)
        vstack.addWidget(self.curveletters)

        self.buttonbox = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel)
        vstack.addWidget(self.buttonbox)

        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.rejected.connect(self.reject)

        self.setLayout(vstack)

    @property
    def result(self):
        return self.formula.text()

def askUserValue(param):
    if param.request == 'time':
        value, isok = QtGui.QInputDialog.getText(None, 'Enter time', param.description)
        try:
            value = ureg.Quantity(value)
            if value.dimensionless:
                # Default to second if no unit is specified
                value = ureg.Quantity(value.magnitude, 's')
            value = value.to_base_units().magnitude
        except (DimensionalityError, UndefinedUnitError, ValueError):
            return None
    elif param.request is str:
        value, isok = QtGui.QInputDialog.getText(None, 'Enter value', param.description)
    elif param.request is int:
        value, isok = QtGui.QInputDialog.getInt(None, 'Enter value', param.description)
    elif param.request is float:
        value, isok = QtGui.QInputDialog.getDouble(None, 'Enter value', param.description, decimals=3)
    elif param.request is bool:
        request = ['Yes', 'No']
        tmpvalue, isok = QtGui.QInputDialog.getItem(None, 'Enter value', param.description, request, editable=False)
        value = (tmpvalue == 'Yes')
    elif param.request is datetime:
        dlg = DlgSetDateTime()
        isok = dlg.exec_()
        value = dlg.result
    elif type(param.request) is list:
        value, isok = QtGui.QInputDialog.getItem(None, 'Enter value', param.description, param.request, editable=False)
    else:
        raise TypeError("Unknown request type")

    if isok:
        return value
    else:
        return None

def userConfirm(question: str, title: str = '') -> bool:
    if not title:
        title = question
    reply = QtGui.QMessageBox.question(None, title, question, QtGui.QMessageBox.Yes, QtGui.QMessageBox.No)
    confirmed = (reply == QtGui.QMessageBox.Yes)
    return confirmed

def askFilePath(
    caption: str,
    filename: str = '',
    folder: str = '',
    filter: str = "CSV files (*.csv *.dat)"
) -> Optional[pathlib.Path]:
    if not folder:
        default = pathlib.Path.home()
    else:
        default = pathlib.Path(folder)

    if filename:
        default = pathlib.Path(default, filename)

    filepath = QtGui.QFileDialog.getSaveFileName(
        caption   = caption,
        filter    = filter,
        directory = str(default)
    )
    if type(filepath) is not str:
        # PyQt5 API change
        filepath = filepath[0]
    if not filepath:
        # Cancel pressed
        return None
    filepath = sanitize_filepath(filepath)
    return pathlib.Path(filepath).resolve()

def askDirPath(
    caption: str,
    folder: str = ''
) -> Optional[pathlib.Path]:
    if not folder:
        folder = str(pathlib.Path.home())

    outdirtmp = QtGui.QFileDialog.getExistingDirectory(caption = caption,
                                                       directory = folder)
    if not outdirtmp:
        # Cancel pressed
        return None
    return pathlib.Path(outdirtmp).resolve()



from graphysio.plotwidgets.plotwidget import PlotWidget
from graphysio.plotwidgets.puplot import LoopWidget
from graphysio.plotwidgets.poiselector import POISelectorWidget
from graphysio.plotwidgets.spectrogram import SpectrogramWidget
from graphysio.plotwidgets.tsplot import TSWidget



import itertools

import numpy as np
import pandas as pd

from graphysio.utils import truncatevecs
from graphysio.algorithms.filters import savgol

def findPressureFeet(curve):
    series = curve.series
    samplerate = curve.samplerate

    fstderiv = series.diff().shift(-1)
    sndderiv = fstderiv.diff().shift(-1)

    # Remove deceleration peaks
    sndderiv = sndderiv * (fstderiv > 0)

    def performWindowing(sumcoef = 4, quantcoef = 3):
        # Find pulse rising edge
        winsum   = int(samplerate / sumcoef)
        winquant = int(samplerate * quantcoef)
        sndderivsq  = sndderiv ** 2
        integral = sndderivsq.rolling(window=winsum, center=True).sum()
        thres = integral.rolling(window=winquant).quantile(.7)
        thres = thres.fillna(method='backfill')
        risings = (integral > thres).astype(int)
        risingvar = risings.diff()
        risingStarts, = (risingvar > 0).nonzero()
        risingStops,  = (risingvar < 0).nonzero()
        return (risingStarts, risingStops)

    found = False
    for quantcoef in [3, 2, 1]:
        # Try again with smaller window if we find nothing
        risingStarts, risingStops = performWindowing(quantcoef=quantcoef)
        try:
            risingStops = risingStops[risingStops > risingStarts[0]]
            found = True
            break
        except IndexError:
            continue

    # Last resort: find one foot on the whole series
    if not found:
        risingStarts = [0]
        risingStops  = [len(sndderiv) - 1]

    def locateMaxima():
        for start, stop in zip(risingStarts, risingStops):
            idxstart = sndderiv.index[start]
            idxstop  = sndderiv.index[stop]
            try:
                maximum = sndderiv.loc[idxstart:idxstop].idxmax()
            except ValueError:
                continue
            else:
                yield maximum

    cycleStarts = pd.Index(list(locateMaxima()))
    return cycleStarts

def findFlowCycles(curve):
    series = curve.series
    bincycles = (series > series.min()).astype(int)
    idxstarts, = (bincycles.diff().shift(-1) > 0).nonzero()
    idxstops,  = (bincycles.diff() < 0).nonzero()
    cycleStarts = series.iloc[idxstarts]
    cycleStops  = series.iloc[idxstops]

    # Handle the case where we start within a cycle
    try:
        cycleStops = cycleStops[cycleStops.index > cycleStarts.index[0]]
    except IndexError as e:
        raise TypeError(f'No cycle detected: {e}')

    return (cycleStarts.index, cycleStops.index)

def findPressureFullBak(curve):
    series = curve.series
    samplerate = curve.samplerate

    fstderivraw = series.diff().iloc[1:]
    sndderivraw = fstderivraw.diff().iloc[1:]
    # Smoothen the derivatives
    fstderiv, _ = savgol(fstderivraw, samplerate, (.16, 2))
    sndderiv, _ = savgol(sndderivraw, samplerate, (.16, 2))

    cycles = []
    starts, durations = curve.getCycleIndices()
    for start, duration in zip(starts, durations):
        stop = start + duration
        diastop = start - duration
        dia = findPOI(series, [start, diastop], 'min', windowsize=.05, forcesign=False)
        sbp = findPOI(series, [start, stop], 'max', windowsize=.05)
        peridic = findPOI(sndderiv, [sbp, stop], 'max', windowsize=.15)
        dic = findHorizontal(fstderiv, peridic)
        cycle = (dia, sbp, dic)
        cycles.append(cycle)

    indices = [pd.Index(idx, dtype=np.int64) for idx in zip(*cycles)]
    return indices

def findPressureCycles(curve):
    series = curve.series

    cycles = []
    starts, durations = curve.getCycleIndices()
    for start, duration in zip(starts, durations):
        stop = start + duration
        diastop = start - duration
        dia = findPOI(series, [start, diastop], 'min', windowsize=.05, forcesign=False)
        sbp = findPOI(series, [start, stop], 'max', windowsize=.05)
        cycle = (dia, sbp)
        cycles.append(cycle)
    indices = [pd.Index(idx, dtype=np.int64) for idx in zip(*cycles)]
    return indices

def findPressureFull(curve):
    dia, sbp = findPressureCycles(curve)
    upstroke_duration = np.abs(sbp - dia)
    dia1, sbp1 = truncatevecs([dia[1:], sbp])
    dic = findDicProj(curve.series, dia1, sbp1, upstroke_duration)
    return [dia, sbp, dic]

# Utility function for point placing

def isbetter(new, ref, kind, forcesign):
    if kind == 'max':
        condition = (new > ref)
        if forcesign:
            condition = condition or (new < 0)
    elif kind == 'min':
        condition = (new < ref)
        if forcesign:
            condition = condition or (new > 0)
    else:
        raise ValueError(kind)
    return condition

def genWindows(soi, interval, windowspan):
    begin, end = interval
    ltr = (end > begin)
    windowspan *= 1e9 # s to ns
    if begin is None or end is None:
        return
    if ltr:
        direction = 1
    else:
        direction = -1
    for n in itertools.count():
        start = begin + direction * n * windowspan
        stop = start + direction * windowspan

        # Stop condition if we exceed end
        if ltr:
            if stop >= end:
                stop = end
        else:
            if stop <= end:
                stop = end
            start, stop = (stop, start)

        window = soi.loc[start:stop]
        if len(window) < 1:
            return
        yield window.index.values

def findPOI(soi, interval, kind, windowsize, forcesign=True):
    if kind not in ['min', 'max']:
        raise ValueError(kind)
    argkind = 'idx' + kind

    goodwindow = []
    previous = -np.inf if kind == 'max' else np.inf
    for window in genWindows(soi, interval, windowsize):
        zoi = soi.loc[window]
        new = getattr(zoi, kind)()
        if isbetter(new, previous, kind, forcesign):
            goodwindow = window
        else:
            break
        previous = new
    finalzoi = soi.loc[goodwindow]
    try:
        retidx = getattr(finalzoi, argkind)()
    except ValueError:
        # No POI found
        retidx = None
    return retidx

def findPOIGreedy(soi, start, kind):
    if kind not in ['min', 'max']:
        raise ValueError(kind)
    loc = soi.index.get_loc(start, method='nearest')
    # Find direction
    try:
        finddir = soi.iloc[[loc-1, loc, loc+1]]
    except IndexError:
        # We're at the edge of the curve
        return start
    npminmax = np.argmin if kind == 'min' else np.argmax
    direction = npminmax(finddir.values) - 1
    if direction == 0:
        # We're already at the local minimum
        return start
    curloc = loc
    while True:
        nextloc = curloc + direction
        try:
            samplesoi = soi.iloc[[curloc, nextloc]]
        except IndexError:
            # We're at the edge of the curve
            break
        nextisbetter = npminmax(samplesoi.values)
        if not nextisbetter:
            # Minimum found
            break
        curloc = nextloc
    return soi.index[curloc]

def findHorizontal(soi, loc):
    if loc is None:
        return None
    step = 8000000 # 8 ms (from ns)
    end = loc + 10 * step
    zoi = soi.loc[loc:end]
    horidx = zoi.abs().idxmin()
    return horidx

def distance(l1, l2, p):
    return np.cross(l2-l1, p-l1) / np.linalg.norm(l2-l1)

def findDicProj(series, dia, sbp, upstroke_duration):
    dics = []
    for si, di, up in zip(sbp, dia, upstroke_duration):
        zoi = series.loc[si:di]
        if len(zoi) < 1:
            continue
        p1 = np.array([si, zoi.iloc[0]])
        p2 = np.array([di, zoi.iloc[-1]])
        p3 = np.vstack([zoi.index, zoi.values]).transpose()
        d = distance(p1, p2, p3)
        # Limit search zone to the beginning of the segment
        search_len = len(series.loc[si:si+2*up])
        search_zone = d[0:search_len]
        argmin = np.argmin(search_zone)
        dics.append(zoi.index[argmin])
    return pd.Index(dics)